@@ -4,10 +4,11 @@ from collections import defaultdict
 import time
 import logging
 import os
-import urllib.request, urllib.parse, urllib.error
 import json
+import warnings
 from datetime import datetime
 from decimal import Decimal
+from itertools import chain, repeat
 from threading import RLock
 
 try:
@@ -16,6 +17,7 @@ try:
 except ImportError:
     pass
 from psycopg2 import connect, Binary
+from psycopg2.sql import SQL, Identifier
 from psycopg2.pool import ThreadedConnectionPool, PoolError
 from psycopg2.extensions import cursor
 from psycopg2.extensions import ISOLATION_LEVEL_REPEATABLE_READ
@@ -28,14 +30,17 @@ except ImportError:
     PYDATE, PYDATETIME, PYTIME, PYINTERVAL = None, None, None, None
 from psycopg2 import IntegrityError as DatabaseIntegrityError
 from psycopg2 import OperationalError as DatabaseOperationalError
+from psycopg2 import ProgrammingError
 from psycopg2.extras import register_default_json, register_default_jsonb
 
-from sql import Flavor
+from sql import Flavor, Cast, For, Table
+from sql.conditionals import Coalesce
 from sql.functions import Function
+from sql.operators import BinaryOperator, Concat
 
 from trytond.backend.database import DatabaseInterface, SQLType
 from trytond.config import config, parse_uri
-from trytond.protocols.jsonrpc import JSONDecoder
+from trytond.tools.gevent import is_gevent_monkey_patched
 
 __all__ = ['Database', 'DatabaseIntegrityError', 'DatabaseOperationalError']
 
@@ -45,6 +50,7 @@ os.environ['PGTZ'] = os.environ.get('TZ', '')
 _timeout = config.getint('database', 'timeout')
 _minconn = config.getint('database', 'minconn', default=1)
 _maxconn = config.getint('database', 'maxconn', default=64)
+_default_name = config.get('database', 'default_name', default='template1')
 
 
 def unescape_quote(s):
@@ -66,11 +72,66 @@ class LoggingCursor(cursor):
         cursor.execute(self, sql, args)
 
 
+class ForSkipLocked(For):
+    def __str__(self):
+        assert not self.nowait, "Can not use both NO WAIT and SKIP LOCKED"
+        return super().__str__() + (' SKIP LOCKED' if not self.nowait else '')
+
+
 class Unaccent(Function):
     __slots__ = ()
     _function = 'unaccent'
 
 
+class Similarity(Function):
+    __slots__ = ()
+    _function = 'similarity'
+
+
+class Match(BinaryOperator):
+    __slots__ = ()
+    _operator = '@@'
+
+
+class ToTsvector(Function):
+    __slots__ = ()
+    _function = 'to_tsvector'
+
+
+class Setweight(Function):
+    __slots__ = ()
+    _function = 'setweight'
+
+
+class TsQuery(Function):
+    __slots__ = ()
+
+
+class ToTsQuery(TsQuery):
+    __slots__ = ()
+    _function = 'to_tsquery'
+
+
+class PlainToTsQuery(TsQuery):
+    __slots__ = ()
+    _function = 'plainto_tsquery'
+
+
+class PhraseToTsQuery(TsQuery):
+    __slots__ = ()
+    _function = 'phraseto_tsquery'
+
+
+class WebsearchToTsQuery(TsQuery):
+    __slots__ = ()
+    _function = 'websearch_to_tsquery'
+
+
+class TsRank(Function):
+    __slots__ = ()
+    _function = 'ts_rank'
+
+
 class AdvisoryLock(Function):
     _function = 'pg_advisory_xact_lock'
 
@@ -79,6 +140,48 @@ class TryAdvisoryLock(Function):
     _function = 'pg_try_advisory_xact_lock'
 
 
+class JSONBExtractPath(Function):
+    __slots__ = ()
+    _function = 'jsonb_extract_path'
+
+
+class JSONKeyExists(BinaryOperator):
+    __slots__ = ()
+    _operator = '?'
+
+
+class _BinaryOperatorArray(BinaryOperator):
+    "Binary Operator that convert list into Array"
+
+    @property
+    def _operands(self):
+        if isinstance(self.right, list):
+            return (self.left, None)
+        return super()._operands
+
+    @property
+    def params(self):
+        params = super().params
+        if isinstance(self.right, list):
+            params = params[:-1] + (self.right,)
+        return params
+
+
+class JSONAnyKeyExist(_BinaryOperatorArray):
+    __slots__ = ()
+    _operator = '?|'
+
+
+class JSONAllKeyExist(_BinaryOperatorArray):
+    __slots__ = ()
+    _operator = '?&'
+
+
+class JSONContains(BinaryOperator):
+    __slots__ = ()
+    _operator = '@>'
+
+
 class Database(DatabaseInterface):
 
     _lock = RLock()
@@ -89,7 +192,9 @@ class Database(DatabaseInterface):
     _search_path = None
     _current_user = None
     _has_returning = None
-    _has_unaccent = {}
+    _has_select_for_skip_locked = None
+    _has_proc = defaultdict(dict)
+    _search_full_text_languages = defaultdict(dict)
     flavor = Flavor(ilike=True)
 
     TYPES_MAPPING = {
@@ -99,9 +204,10 @@ class Database(DatabaseInterface):
         'BLOB': SQLType('BYTEA', 'BYTEA'),
         'DATETIME': SQLType('TIMESTAMP', 'TIMESTAMP(0)'),
         'TIMESTAMP': SQLType('TIMESTAMP', 'TIMESTAMP(6)'),
+        'FULLTEXT': SQLType('TSVECTOR', 'TSVECTOR'),
         }
 
-    def __new__(cls, name='template1'):
+    def __new__(cls, name=_default_name):
         with cls._lock:
             now = datetime.now()
             databases = cls._databases[os.getpid()]
@@ -113,37 +219,36 @@ class Database(DatabaseInterface):
             if name in databases:
                 inst = databases[name]
             else:
-                if name == 'template1':
+                if name == _default_name:
                     minconn = 0
                 else:
                     minconn = _minconn
                 inst = DatabaseInterface.__new__(cls, name=name)
-                logger.info('connect to "%s"', name)
-                inst._connpool = ThreadedConnectionPool(
-                    minconn, _maxconn,
-                    cursor_factory=LoggingCursor,
-                    **cls._connection_params(name))
+                try:
+                    inst._connpool = ThreadedConnectionPool(
+                        minconn, _maxconn, **cls._connection_params(name),
+                        cursor_factory=LoggingCursor)
+                except Exception:
+                    logger.error(
+                        'connection to "%s" failed', name, exc_info=True)
+                    raise
+                else:
+                    logger.info('connection to "%s" succeeded', name)
                 databases[name] = inst
             inst._last_use = datetime.now()
             return inst
 
-    def __init__(self, name='template1'):
+    def __init__(self, name=_default_name):
         super(Database, self).__init__(name)
 
     @classmethod
     def _connection_params(cls, name):
         uri = parse_uri(config.get('database', 'uri'))
+        if uri.path and uri.path != '/':
+            warnings.warn("The path specified in the URI will be overridden")
         params = {
-            'dbname': name,
+            'dsn': uri._replace(path=name).geturl(),
             }
-        if uri.username:
-            params['user'] = uri.username
-        if uri.password:
-            params['password'] = urllib.parse.unquote_plus(uri.password)
-        if uri.hostname:
-            params['host'] = uri.hostname
-        if uri.port:
-            params['port'] = uri.port
         return params
 
     def connect(self):
@@ -160,6 +265,10 @@ class Database(DatabaseInterface):
                     time.sleep(1)
                     continue
                 raise
+            except Exception:
+                logger.error(
+                    'connection to "%s" failed', self.name, exc_info=True)
+                raise
         # We do not use set_session because psycopg2 < 2.7 and psycopg2cffi
         # change the default_transaction_* attributes which breaks external
         # pooling at the transaction level.
@@ -180,21 +289,26 @@ class Database(DatabaseInterface):
 
     def close(self):
         with self._lock:
-            logger.info('disconnect from "%s"', self.name)
+            logger.info('disconnection from "%s"', self.name)
             self._connpool.closeall()
             self._databases[os.getpid()].pop(self.name)
 
     @classmethod
     def create(cls, connection, database_name, template='template0'):
         cursor = connection.cursor()
-        cursor.execute('CREATE DATABASE "' + database_name + '" '
-            'TEMPLATE "' + template + '" ENCODING \'unicode\'')
+        cursor.execute(
+            SQL(
+                "CREATE DATABASE {} TEMPLATE {} ENCODING 'unicode'")
+            .format(
+                Identifier(database_name),
+                Identifier(template)))
         connection.commit()
         cls._list_cache.clear()
 
     def drop(self, connection, database_name):
         cursor = connection.cursor()
-        cursor.execute('DROP DATABASE "' + database_name + '"')
+        cursor.execute(SQL("DROP DATABASE {}")
+            .format(Identifier(database_name)))
         self.__class__._list_cache.clear()
 
     def get_version(self, connection):
@@ -219,11 +333,16 @@ class Database(DatabaseInterface):
             res = []
             for db_name, in cursor:
                 try:
-                    with connect(**self._connection_params(db_name)
-                            ) as conn:
-                        if self._test(conn, hostname=hostname):
-                            res.append(db_name)
+                    conn = connect(**self._connection_params(db_name))
+                    try:
+                        with conn:
+                            if self._test(conn, hostname=hostname):
+                                res.append(db_name)
+                    finally:
+                        conn.close()
                 except Exception:
+                    logger.debug(
+                        'Test failed for "%s"', db_name, exc_info=True)
                     continue
         finally:
             self.put_connection(connection)
@@ -266,14 +385,15 @@ class Database(DatabaseInterface):
             self.put_connection(connection)
 
     def test(self, hostname=None):
-        connection = self.get_connection()
         try:
-            is_tryton_database = self._test(connection, hostname=hostname)
+            connection = self.get_connection()
         except Exception:
-            is_tryton_database = False
+            logger.debug('Test failed for "%s"', self.name, exc_info=True)
+            return False
+        try:
+            return self._test(connection, hostname=hostname)
         finally:
             self.put_connection(connection)
-        return is_tryton_database
 
     @classmethod
     def _test(cls, connection, hostname=None):
@@ -286,30 +406,35 @@ class Database(DatabaseInterface):
         if len(cursor.fetchall()) != len(tables):
             return False
         if hostname:
-            cursor.execute(
-                'SELECT hostname FROM ir_configuration')
-            hostnames = {h for h, in cursor.fetchall() if h}
-            if hostnames and hostname not in hostnames:
-                return False
+            try:
+                cursor.execute(
+                    'SELECT hostname FROM ir_configuration')
+                hostnames = {h for h, in cursor if h}
+                if hostnames and hostname not in hostnames:
+                    return False
+            except ProgrammingError:
+                pass
         return True
 
     def nextid(self, connection, table):
         cursor = connection.cursor()
-        cursor.execute("SELECT NEXTVAL('" + table + "_id_seq')")
+        cursor.execute("SELECT NEXTVAL(%s)", (table + '_id_seq',))
         return cursor.fetchone()[0]
 
     def setnextid(self, connection, table, value):
         cursor = connection.cursor()
-        cursor.execute("SELECT SETVAL('" + table + "_id_seq', %d)" % value)
+        cursor.execute("SELECT SETVAL(%s, %s)", (table + '_id_seq', value))
 
     def currid(self, connection, table):
         cursor = connection.cursor()
-        cursor.execute('SELECT last_value FROM "' + table + '_id_seq"')
+        cursor.execute(SQL("SELECT last_value FROM {}").format(
+                Identifier(table + '_id_seq')))
         return cursor.fetchone()[0]
 
     def lock(self, connection, table):
         cursor = connection.cursor()
-        cursor.execute('LOCK "%s" IN EXCLUSIVE MODE NOWAIT' % table)
+        cursor.execute(SQL('LOCK {} IN EXCLUSIVE MODE NOWAIT').format(
+                Identifier(table)))
 
     def lock_id(self, id, timeout=None):
         if not timeout:
@@ -377,6 +502,20 @@ class Database(DatabaseInterface):
     def has_select_for(self):
         return True
 
+    def get_select_for_skip_locked(self):
+        if self._has_select_for_skip_locked is None:
+            connection = self.get_connection()
+            try:
+                # SKIP LOCKED clause is available since PostgreSQL 9.5
+                self._has_select_for_skip_locked = (
+                    self.get_version(connection) >= (9, 5))
+            finally:
+                self.put_connection(connection)
+        if self._has_select_for_skip_locked:
+            return ForSkipLocked
+        else:
+            return For
+
     def has_window_functions(self):
         return True
 
@@ -384,21 +523,117 @@ class Database(DatabaseInterface):
     def has_sequence(cls):
         return True
 
-    def has_unaccent(self):
-        if self.name in self._has_unaccent:
-            return self._has_unaccent[self.name]
+    def has_proc(self, name):
+        if name in self._has_proc[self.name]:
+            return self._has_proc[self.name][name]
         connection = self.get_connection()
-        unaccent = False
+        result = False
         try:
             cursor = connection.cursor()
             cursor.execute(
                 "SELECT 1 FROM pg_proc WHERE proname=%s",
-                (Unaccent._function,))
-            unaccent = bool(cursor.rowcount)
+                (name,))
+            result = bool(cursor.rowcount)
         finally:
             self.put_connection(connection)
-        self._has_unaccent[self.name] = unaccent
-        return unaccent
+        self._has_proc[self.name][name] = result
+        return result
+
+    def has_unaccent(self):
+        return self.has_proc(Unaccent._function)
+
+    def has_similarity(self):
+        return self.has_proc(Similarity._function)
+
+    def similarity(self, column, value):
+        return Similarity(column, value)
+
+    def has_search_full_text(self):
+        return True
+
+    def _search_full_text_language(self, language):
+        languages = self._search_full_text_languages[self.name]
+        if language not in languages:
+            lang = Table('ir_lang')
+            connection = self.get_connection()
+            try:
+                cursor = connection.cursor()
+                cursor.execute(*lang.select(
+                        Coalesce(lang.pg_text_search, 'simple'),
+                        where=lang.code == language,
+                        limit=1))
+                config_name, = cursor.fetchone()
+            finally:
+                self.put_connection(connection)
+            languages[language] = config_name
+        else:
+            config_name = languages[language]
+        return config_name
+
+    def format_full_text(self, *documents, language=None):
+        size = max(len(documents) // 4, 1)
+        if len(documents) > 1:
+            weights = chain(
+                ['A'] * size, ['B'] * size, ['C'] * size, repeat('D'))
+        else:
+            weights = [None]
+        expression = None
+        if language:
+            config_name = self._search_full_text_language(language)
+        else:
+            config_name = None
+        for document, weight in zip(documents, weights):
+            if not document:
+                continue
+            if config_name:
+                ts_vector = ToTsvector(config_name, document)
+            else:
+                ts_vector = ToTsvector('simple', document)
+            if weight:
+                ts_vector = Setweight(ts_vector, weight)
+            if expression is None:
+                expression = ts_vector
+            else:
+                expression = Concat(expression, ts_vector)
+        return expression
+
+    def format_full_text_query(self, query, language=None):
+        connection = self.get_connection()
+        try:
+            version = self.get_version(connection)
+        finally:
+            self.put_connection(connection)
+        if version >= (11, 0):
+            ToTsQuery = WebsearchToTsQuery
+        else:
+            ToTsQuery = PlainToTsQuery
+        if language:
+            config_name = self._search_full_text_language(language)
+            if not isinstance(query, TsQuery):
+                query = ToTsQuery(config_name, query)
+        else:
+            if not isinstance(query, TsQuery):
+                query = ToTsQuery(query)
+        return query
+
+    def search_full_text(self, document, query):
+        return Match(document, query)
+
+    def rank_full_text(self, document, query, normalize=None):
+        # TODO: weights and cover density
+        norm_int = 0
+        if normalize:
+            values = {
+                'document log': 1,
+                'document': 2,
+                'mean': 4,
+                'word': 8,
+                'word log': 16,
+                'rank': 32,
+                }
+            for norm in normalize:
+                norm_int |= values.get(norm, 0)
+        return TsRank(document, query, norm_int)
 
     def sql_type(self, type_):
         if type_ in self.TYPES_MAPPING:
@@ -433,35 +668,32 @@ class Database(DatabaseInterface):
             self, connection, name, number_increment=1, start_value=1):
         cursor = connection.cursor()
 
-        param = self.flavor.param
         cursor.execute(
-            'CREATE SEQUENCE "%s" '
-            'INCREMENT BY %s '
-            'START WITH %s'
-            % (name, param, param),
+            SQL("CREATE SEQUENCE {} INCREMENT BY %s START WITH %s").format(
+                Identifier(name)),
             (number_increment, start_value))
 
     def sequence_update(
             self, connection, name, number_increment=1, start_value=1):
         cursor = connection.cursor()
-        param = self.flavor.param
         cursor.execute(
-            'ALTER SEQUENCE "%s" '
-            'INCREMENT BY %s '
-            'RESTART WITH %s'
-            % (name, param, param),
+            SQL("ALTER SEQUENCE {} INCREMENT BY %s RESTART WITH %s").format(
+                Identifier(name)),
             (number_increment, start_value))
 
     def sequence_rename(self, connection, old_name, new_name):
         cursor = connection.cursor()
         if (self.sequence_exist(connection, old_name)
                 and not self.sequence_exist(connection, new_name)):
-            cursor.execute('ALTER TABLE "%s" RENAME TO "%s"'
-                % (old_name, new_name))
+            cursor.execute(
+                SQL("ALTER TABLE {} RENAME TO {}").format(
+                    Identifier(old_name),
+                    Identifier(new_name)))
 
     def sequence_delete(self, connection, name):
         cursor = connection.cursor()
-        cursor.execute('DROP SEQUENCE "%s"' % name)
+        cursor.execute(SQL("DROP SEQUENCE {}").format(
+                Identifier(name)))
 
     def sequence_next_number(self, connection, name):
         cursor = connection.cursor()
@@ -470,27 +702,46 @@ class Database(DatabaseInterface):
             cursor.execute(
                 'SELECT increment_by '
                 'FROM pg_sequences '
-                'WHERE sequencename=%s '
-                % self.flavor.param,
+                'WHERE sequencename=%s',
                 (name,))
             increment, = cursor.fetchone()
             cursor.execute(
-                'SELECT CASE WHEN NOT is_called THEN last_value '
-                            'ELSE last_value + %s '
-                        'END '
-                'FROM "%s"' % (self.flavor.param, name),
+                SQL(
+                    'SELECT CASE WHEN NOT is_called THEN last_value '
+                    'ELSE last_value + %s '
+                    'END '
+                    'FROM {}').format(Identifier(name)),
                 (increment,))
         else:
             cursor.execute(
-                'SELECT CASE WHEN NOT is_called THEN last_value '
-                            'ELSE last_value + increment_by '
-                       'END '
-                'FROM "%s"' % name)
+                SQL(
+                    'SELECT CASE WHEN NOT is_called THEN last_value '
+                    'ELSE last_value + increment_by '
+                    'END '
+                    'FROM {}').format(Identifier(name)))
         return cursor.fetchone()[0]
 
     def has_channel(self):
         return True
 
+    def json_get(self, column, key=None):
+        column = Cast(column, 'jsonb')
+        if key:
+            column = JSONBExtractPath(column, key)
+        return column
+
+    def json_key_exists(self, column, key):
+        return JSONKeyExists(Cast(column, 'jsonb'), key)
+
+    def json_any_keys_exist(self, column, keys):
+        return JSONAnyKeyExist(Cast(column, 'jsonb'), keys)
+
+    def json_all_keys_exist(self, column, keys):
+        return JSONAllKeyExist(Cast(column, 'jsonb'), keys)
+
+    def json_contains(self, column, json):
+        return JSONContains(Cast(column, 'jsonb'), Cast(json, 'jsonb'))
+
 
 register_type(UNICODE)
 if PYDATE:
@@ -506,6 +757,14 @@ register_adapter(Decimal, lambda value: AsIs(str(value)))
 
 
 def convert_json(value):
+    from trytond.protocols.jsonrpc import JSONDecoder
     return json.loads(value, object_hook=JSONDecoder())
+
+
 register_default_json(loads=convert_json)
 register_default_jsonb(loads=convert_json)
+
+if is_gevent_monkey_patched():
+    from psycopg2.extensions import set_wait_callback
+    from psycopg2.extras import wait_select
+    set_wait_callback(wait_select)
