@@ -1,33 +1,40 @@
 # This file is part of Tryton.  The COPYRIGHT file at the top level of
 # this repository contains the full copyright notices and license terms.
-from trytond.backend.database import DatabaseInterface, SQLType
-from trytond.config import config
-import os
-from decimal import Decimal
 import datetime
-import time
-import sys
-import threading
 import logging
+import math
+import os
+import random
+import threading
+import time
+import urllib.parse
+import warnings
+from decimal import Decimal
+from weakref import WeakKeyDictionary
+from werkzeug.security import safe_join
 
-_FIX_ROWCOUNT = False
 try:
     from pysqlite2 import dbapi2 as sqlite
     from pysqlite2.dbapi2 import IntegrityError as DatabaseIntegrityError
     from pysqlite2.dbapi2 import OperationalError as DatabaseOperationalError
-    # pysqlite2 < 2.5 doesn't return correct rowcount
-    _FIX_ROWCOUNT = sqlite.version_info < (2, 5, 0)
 except ImportError:
     import sqlite3 as sqlite
     from sqlite3 import IntegrityError as DatabaseIntegrityError
     from sqlite3 import OperationalError as DatabaseOperationalError
-from sql import Flavor, Table, Query, Expression
+from sql import Flavor, Table, Query, Expression, Literal, Null
+from sql.conditionals import NullIf
 from sql.functions import (Function, Extract, Position, Substring,
     Overlay, CharLength, CurrentTimestamp, Trim)
 
+from trytond.backend.database import DatabaseInterface, SQLType
+from trytond.config import config, parse_uri
+from trytond.transaction import Transaction
+
 __all__ = ['Database', 'DatabaseIntegrityError', 'DatabaseOperationalError']
 logger = logging.getLogger(__name__)
 
+_default_name = config.get('database', 'default_name', default=':memory:')
+
 
 class SQLiteExtract(Function):
     __slots__ = ()
@@ -75,19 +82,33 @@ class SQLiteExtract(Function):
 
 
 def date_trunc(_type, date):
-    if _type == 'second':
+    if not _type:
         return date
-    try:
-        tm_tuple = time.strptime(date, '%Y-%m-%d %H:%M:%S')
-    except Exception:
+    for format_ in [
+            '%Y-%m-%d %H:%M:%S.%f',
+            '%Y-%m-%d %H:%M:%S',
+            '%Y-%m-%d',
+            '%H:%M:%S',
+            ]:
+        try:
+            value = datetime.datetime.strptime(date, format_)
+        except ValueError:
+            continue
+        else:
+            break
+    else:
         return None
-    if _type == 'year':
-        return "%i-01-01 00:00:00" % tm_tuple.tm_year
-    elif _type == 'month':
-        return "%i-%02i-01 00:00:00" % (tm_tuple.tm_year, tm_tuple.tm_mon)
-    elif _type == 'day':
-        return "%i-%02i-%02i 00:00:00" % (tm_tuple.tm_year, tm_tuple.tm_mon,
-                tm_tuple.tm_mday)
+    for attribute, replace in [
+            ('microsecond', 0),
+            ('second', 0),
+            ('minute', 0),
+            ('hour', 0),
+            ('day', 1),
+            ('month', 1)]:
+        if _type.startswith(attribute):
+            break
+        value = value.replace(**{attribute: replace})
+    return str(value)
 
 
 def split_part(text, delimiter, count):
@@ -115,7 +136,56 @@ def replace(text, pattern, replacement):
 
 
 def now():
-    return datetime.datetime.now().isoformat(' ')
+    transaction = Transaction()
+    return _nows.setdefault(transaction, {}).setdefault(
+        transaction.started_at, datetime.datetime.now().isoformat(' '))
+
+
+_nows = WeakKeyDictionary()
+
+
+def to_char(value, format):
+    try:
+        value = datetime.datetime.strptime(value, '%Y-%m-%d %H:%M:%S.%f')
+    except ValueError:
+        try:
+            value = datetime.datetime.strptime(value, '%Y-%m-%d').date()
+        except ValueError:
+            pass
+    if isinstance(value, datetime.date):
+        # Convert SQL pattern into compatible Python
+        return value.strftime(format
+            .replace('%', '%%')
+            .replace('HH12', '%I')
+            .replace('HH24', '%H')
+            .replace('HH', '%I')
+            .replace('MI', '%M')
+            .replace('SS', '%S')
+            .replace('US', '%f')
+            .replace('AM', '%p')
+            .replace('A.M.', '%p')
+            .replace('PM', '%p')
+            .replace('P.M.', '%p')
+            .replace('am', '%p')
+            .replace('a.m.', '%p')
+            .replace('pm', '%p')
+            .replace('p.m.', '%p')
+            .replace('YYYY', '%Y')
+            .replace('YY', '%y')
+            .replace('Month', '%B')
+            .replace('Mon', '%b')
+            .replace('MM', '%m')
+            .replace('Day', '%A')
+            .replace('Dy', '%a')
+            .replace('DDD', '%j')
+            .replace('DD', '%d')
+            .replace('D', '%w')
+            .replace('TZ', '%Z')
+            )
+    elif isinstance(value, datetime.timedelta):
+        raise NotImplementedError
+    else:
+        raise NotImplementedError
 
 
 class SQLiteSubstring(Function):
@@ -157,13 +227,22 @@ class SQLiteTrim(Trim):
             }[self.position]
 
         def format(arg):
-            if isinstance(arg, basestring):
+            if isinstance(arg, str):
                 return param
             else:
                 return str(arg)
         return function + '(%s, %s)' % (
             format(self.string), format(self.characters))
 
+    @property
+    def params(self):
+        if isinstance(self.string, str):
+            params = [self.string]
+        else:
+            params = list(self.string.params)
+        params.append(self.characters)
+        return params
+
 
 def sign(value):
     if value > 0:
@@ -175,7 +254,7 @@ def sign(value):
 
 
 def greatest(*args):
-    args = filter(lambda a: a is not None, args)
+    args = [a for a in args if a is not None]
     if args:
         return max(args)
     else:
@@ -183,13 +262,25 @@ def greatest(*args):
 
 
 def least(*args):
-    args = filter(lambda a: a is not None, args)
+    args = [a for a in args if a is not None]
     if args:
         return min(args)
     else:
         return None
 
 
+def cbrt(value):
+    return math.pow(value, 1 / 3)
+
+
+def div(a, b):
+    return a // b
+
+
+def trunc(value, digits):
+    return math.trunc(value * 10 ** digits) / 10 ** digits
+
+
 MAPPING = {
     Extract: SQLiteExtract,
     Position: SQLitePosition,
@@ -201,6 +292,16 @@ MAPPING = {
     }
 
 
+class JSONExtract(Function):
+    __slots__ = ()
+    _function = 'JSON_EXTRACT'
+
+
+class JSONQuote(Function):
+    __slots__ = ()
+    _function = 'JSON_QUOTE'
+
+
 class SQLiteCursor(sqlite.Cursor):
 
     def __enter__(self):
@@ -228,50 +329,105 @@ class Database(DatabaseInterface):
         'DATETIME': SQLType('TIMESTAMP', 'TIMESTAMP'),
         'BIGINT': SQLType('INTEGER', 'INTEGER'),
         'BOOL': SQLType('BOOLEAN', 'BOOLEAN'),
+        'FULLTEXT': SQLType('TEXT', 'TEXT'),
         }
 
-    def __new__(cls, name=':memory:'):
+    def __new__(cls, name=_default_name):
         if (name == ':memory:'
                 and getattr(cls._local, 'memory_database', None)):
             return cls._local.memory_database
         return DatabaseInterface.__new__(cls, name=name)
 
-    def __init__(self, name=':memory:'):
+    def __init__(self, name=_default_name):
         super(Database, self).__init__(name=name)
         if name == ':memory:':
             Database._local.memory_database = self
 
     def connect(self):
-        if self.name == ':memory:':
-            path = ':memory:'
-        else:
-            db_filename = self.name + '.sqlite'
-            path = os.path.join(config.get('database', 'path'), db_filename)
-            if not os.path.isfile(path):
-                raise IOError('Database "%s" doesn\'t exist!' % db_filename)
         if self._conn is not None:
             return self
-        self._conn = sqlite.connect(path,
+        self._conn = sqlite.connect(
+            self._make_uri(), uri=True,
             detect_types=sqlite.PARSE_DECLTYPES | sqlite.PARSE_COLNAMES,
             factory=SQLiteConnection)
         self._conn.create_function('extract', 2, SQLiteExtract.extract)
         self._conn.create_function('date_trunc', 2, date_trunc)
         self._conn.create_function('split_part', 3, split_part)
-        self._conn.create_function('position', 2, SQLitePosition.position)
-        self._conn.create_function('overlay', 3, SQLiteOverlay.overlay)
-        self._conn.create_function('overlay', 4, SQLiteOverlay.overlay)
+        self._conn.create_function('to_char', 2, to_char)
         if sqlite.sqlite_version_info < (3, 3, 14):
             self._conn.create_function('replace', 3, replace)
         self._conn.create_function('now', 0, now)
-        self._conn.create_function('sign', 1, sign)
         self._conn.create_function('greatest', -1, greatest)
         self._conn.create_function('least', -1, least)
+
+        # Mathematical functions
+        self._conn.create_function('cbrt', 1, cbrt)
+        self._conn.create_function('ceil', 1, math.ceil)
+        self._conn.create_function('degrees', 1, math.degrees)
+        self._conn.create_function('div', 2, div)
+        self._conn.create_function('exp', 1, math.exp)
+        self._conn.create_function('floor', 1, math.floor)
+        self._conn.create_function('ln', 1, math.log)
+        self._conn.create_function('log', 1, math.log10)
+        self._conn.create_function('mod', 2, math.fmod)
+        self._conn.create_function('pi', 0, lambda: math.pi)
+        self._conn.create_function('power', 2, math.pow)
+        self._conn.create_function('radians', 1, math.radians)
+        self._conn.create_function('sign', 1, sign)
+        self._conn.create_function('sqrt', 1, math.sqrt)
+        self._conn.create_function('trunc', 1, math.trunc)
+        self._conn.create_function('trunc', 2, trunc)
+
+        # Trigonomentric functions
+        self._conn.create_function('acos', 1, math.acos)
+        self._conn.create_function('asin', 1, math.asin)
+        self._conn.create_function('atan', 1, math.atan)
+        self._conn.create_function('atan2', 2, math.atan2)
+        self._conn.create_function('cos', 1, math.cos)
+        self._conn.create_function(
+            'cot', 1, lambda x: 1 / math.tan(x) if x else math.inf)
+        self._conn.create_function('sin', 1, math.sin)
+        self._conn.create_function('tan', 1, math.tan)
+
+        # Random functions
+        self._conn.create_function('random', 0, random.random)
+        self._conn.create_function('setseed', 1, random.seed)
+
+        # String functions
+        self._conn.create_function('overlay', 3, SQLiteOverlay.overlay)
+        self._conn.create_function('overlay', 4, SQLiteOverlay.overlay)
+        self._conn.create_function('position', 2, SQLitePosition.position)
+
         if (hasattr(self._conn, 'set_trace_callback')
                 and logger.isEnabledFor(logging.DEBUG)):
             self._conn.set_trace_callback(logger.debug)
         self._conn.execute('PRAGMA foreign_keys = ON')
         return self
 
+    def _make_uri(self):
+        uri = config.get('database', 'uri')
+        base_uri = parse_uri(uri)
+        if base_uri.path and base_uri.path != '/':
+            warnings.warn("The path specified in the URI will be overridden")
+
+        if self.name == ':memory:':
+            query_string = urllib.parse.parse_qs(base_uri.query)
+            query_string['mode'] = 'memory'
+            query = urllib.parse.urlencode(query_string, doseq=True)
+            db_uri = base_uri._replace(netloc='', path='/', query=query)
+        else:
+            db_path = safe_join(
+                config.get('database', 'path'), self.name + '.sqlite')
+            if not os.path.isfile(db_path):
+                raise IOError("Database '%s' doesn't exist!" % db_path)
+            db_uri = base_uri._replace(path=db_path)
+
+        # Use unparse before replacing sqlite with file because SQLite accepts
+        # a relative path URI like file:db/test.sqlite which doesn't conform to
+        # RFC8089 which urllib follows and enforces when the scheme is 'file'
+        db_uri = urllib.parse.urlunparse(db_uri)
+        return db_uri.replace('sqlite', 'file', 1)
+
     def get_connection(self, autocommit=False, readonly=False):
         if self._conn is None:
             self.connect()
@@ -313,7 +469,7 @@ class Database(DatabaseInterface):
         os.remove(os.path.join(config.get('database', 'path'),
             database_name + '.sqlite'))
 
-    def list(self):
+    def list(self, hostname=None):
         res = []
         listdir = [':memory:']
         try:
@@ -329,14 +485,17 @@ class Database(DatabaseInterface):
                 try:
                     database = Database(db_name).connect()
                 except Exception:
+                    logger.debug(
+                        'Test failed for "%s"', db_name, exc_info=True)
                     continue
-                if database.test():
+                if database.test(hostname=hostname):
                     res.append(db_name)
                 database.close()
         return res
 
     def init(self):
         from trytond.modules import get_module_info
+        Flavor.set(self.flavor)
         with self.get_connection() as conn:
             cursor = conn.cursor()
             sql_file = os.path.join(os.path.dirname(__file__), 'init.sql')
@@ -370,29 +529,34 @@ class Database(DatabaseInterface):
                     cursor.execute(*insert)
             conn.commit()
 
-    def test(self):
+    def test(self, hostname=None):
+        Flavor.set(self.flavor)
+        tables = ['ir_model', 'ir_model_field', 'ir_ui_view', 'ir_ui_menu',
+            'res_user', 'res_group', 'ir_module', 'ir_module_dependency',
+            'ir_translation', 'ir_lang', 'ir_configuration']
         sqlite_master = Table('sqlite_master')
         select = sqlite_master.select(sqlite_master.name)
         select.where = sqlite_master.type == 'table'
-        select.where &= sqlite_master.name.in_([
-                'ir_model',
-                'ir_model_field',
-                'ir_ui_view',
-                'ir_ui_menu',
-                'res_user',
-                'res_group',
-                'ir_module',
-                'ir_module_dependency',
-                'ir_translation',
-                'ir_lang',
-                ])
+        select.where &= sqlite_master.name.in_(tables)
         with self._conn as conn:
             cursor = conn.cursor()
             try:
                 cursor.execute(*select)
             except Exception:
                 return False
-            return len(cursor.fetchall()) != 0
+            if len(cursor.fetchall()) != len(tables):
+                return False
+            if hostname:
+                configuration = Table('ir_configuration')
+                try:
+                    cursor.execute(*configuration.select(
+                            configuration.hostname))
+                except Exception:
+                    return False
+                hostnames = {h for h, in cursor if h}
+                if hostnames and hostname not in hostnames:
+                    return False
+        return True
 
     def lastid(self, cursor):
         # This call is not thread safe
@@ -401,7 +565,10 @@ class Database(DatabaseInterface):
     def lock(self, connection, table):
         pass
 
-    def has_constraint(self):
+    def lock_id(self, id, timeout=None):
+        return Literal(True)
+
+    def has_constraint(self, constraint):
         return False
 
     def has_multirow_insert(self):
@@ -421,20 +588,24 @@ class Database(DatabaseInterface):
                 value = int(value)
         return value
 
+    def json_get(self, column, key=None):
+        if key:
+            column = JSONExtract(column, '$.%s' % key)
+        return NullIf(JSONQuote(column), JSONQuote(Null))
+
+
 sqlite.register_converter('NUMERIC', lambda val: Decimal(val.decode('utf-8')))
-if sys.version_info[0] == 2:
-    sqlite.register_adapter(Decimal, lambda val: buffer(str(val)))
-    sqlite.register_adapter(bytearray, lambda val: buffer(val))
-else:
-    sqlite.register_adapter(Decimal, lambda val: str(val).encode('utf-8'))
+sqlite.register_adapter(Decimal, lambda val: str(val).encode('utf-8'))
 
 
 def adapt_datetime(val):
     return val.replace(tzinfo=None).isoformat(" ")
+
+
 sqlite.register_adapter(datetime.datetime, adapt_datetime)
 sqlite.register_adapter(datetime.time, lambda val: val.isoformat())
-sqlite.register_converter('TIME', lambda val: datetime.time(*map(int,
-            val.decode('utf-8').split(':'))))
+sqlite.register_converter('TIME',
+    lambda val: datetime.time(*map(int, val.decode('utf-8').split(':'))))
 sqlite.register_adapter(datetime.timedelta, lambda val: val.total_seconds())
 
 
@@ -446,6 +617,8 @@ def convert_interval(value):
     elif value <= _interval_min:
         return datetime.timedelta.min
     return datetime.timedelta(seconds=value)
+
+
 _interval_max = datetime.timedelta.max.total_seconds()
 _interval_min = datetime.timedelta.min.total_seconds()
 sqlite.register_converter('INTERVAL', convert_interval)
