@@ -1,14 +1,22 @@
 # This file is part of Tryton.  The COPYRIGHT file at the top level of
 # this repository contains the full copyright notices and license terms.
-import os
+import __main__ as main
 import configparser
-import urllib.parse
 import logging
+import os
+import urllib.parse
+
+from . import status
 
 __all__ = ['config', 'get_hostname', 'get_port', 'split_netloc',
     'parse_listen', 'parse_uri']
 logger = logging.getLogger(__name__)
 
+# Needed so urlunsplit to always set netloc
+for backend_name in ['postgresql', 'sqlite']:
+    if backend_name not in urllib.parse.uses_netloc:
+        urllib.parse.uses_netloc.append(backend_name)
+
 
 def get_hostname(netloc):
     if '[' in netloc and ']' in netloc:
@@ -37,52 +45,55 @@ def parse_uri(uri):
     return urllib.parse.urlparse(uri)
 
 
-class TrytonConfigParser(configparser.RawConfigParser):
+class TrytonConfigParser(configparser.ConfigParser):
 
     def __init__(self):
-        configparser.RawConfigParser.__init__(self)
+        super().__init__(interpolation=None)
         self.add_section('web')
         self.set('web', 'listen', 'localhost:8000')
         self.set('web', 'root', os.path.join(os.path.expanduser('~'), 'www'))
-        self.set('web', 'num_proxies', 0)
+        self.set('web', 'num_proxies', '0')
+        self.set('web', 'cache_timeout', str(60 * 60 * 12))
         self.add_section('database')
         self.set('database', 'uri',
             os.environ.get('TRYTOND_DATABASE_URI', 'sqlite://'))
         self.set('database', 'path', os.path.join(
                 os.path.expanduser('~'), 'db'))
         self.set('database', 'list', 'True')
-        self.set('database', 'retry', 5)
+        self.set('database', 'retry', '5')
         self.set('database', 'language', 'en')
-        self.set('database', 'timeout', 30 * 60)
+        self.set('database', 'timeout', str(30 * 60))
         self.add_section('request')
-        self.set('request', 'max_size', 2 * 1024 * 1024)
-        self.set('request', 'max_size_authenticated', 2 * 1024 * 1024 * 1024)
+        self.set('request', 'max_size', str(2 * 1024 * 1024))
+        self.set('request', 'max_size_authenticated',
+            str(2 * 1024 * 1024 * 1024))
         self.add_section('cache')
-        self.set('cache', 'model', 200)
-        self.set('cache', 'record', 2000)
-        self.set('cache', 'field', 100)
+        self.set('cache', 'model', '200')
+        self.set('cache', 'record', '2000')
+        self.set('cache', 'field', '100')
         self.add_section('queue')
-        self.set('queue', 'worker', False)
+        self.set('queue', 'worker', 'False')
         self.add_section('ssl')
         self.add_section('email')
         self.set('email', 'uri', 'smtp://localhost:25')
         self.add_section('session')
         self.set('session', 'authentications', 'password')
-        self.set('session', 'max_age', 60 * 60 * 24 * 30)
-        self.set('session', 'timeout', 60 * 5)
-        self.set('session', 'max_attempt', 5)
-        self.set('session', 'max_attempt_ip_network', 300)
-        self.set('session', 'ip_network_4', 32)
-        self.set('session', 'ip_network_6', 56)
+        self.set('session', 'max_age', str(60 * 60 * 24 * 30))
+        self.set('session', 'timeout', str(60 * 5))
+        self.set('session', 'max_attempt', '5')
+        self.set('session', 'max_attempt_ip_network', '300')
+        self.set('session', 'ip_network_4', '32')
+        self.set('session', 'ip_network_6', '56')
         self.add_section('password')
-        self.set('password', 'length', 8)
-        self.set('password', 'entropy', 0.75)
-        self.set('password', 'reset_timeout', 24 * 60 * 60)
+        self.set('password', 'length', '8')
+        self.set('password', 'entropy', '0.75')
+        self.set('password', 'reset_timeout', str(24 * 60 * 60))
         self.add_section('bus')
-        self.set('bus', 'allow_subscribe', False)
-        self.set('bus', 'long_polling_timeout', 5 * 60)
-        self.set('bus', 'cache_timeout', 5)
-        self.set('bus', 'select_timeout', 5)
+        self.set('bus', 'allow_subscribe', 'False')
+        self.set('bus', 'long_polling_timeout', str(5 * 60))
+        self.set('bus', 'cache_timeout', '5')
+        self.set('bus', 'select_timeout', '5')
+        self.add_section('html')
         self.update_environ()
         self.update_etc()
 
@@ -146,4 +157,8 @@ class TrytonConfigParser(configparser.RawConfigParser):
                 AttributeError):
             return default
 
+
 config = TrytonConfigParser()
+
+if os.path.basename(getattr(main, '__file__', '')) != 'trytond-stat':
+    status.start(config.get('database', 'path'))
