@@ -1,10 +1,8 @@
 # This file is part of Tryton.  The COPYRIGHT file at the top level of
 # this repository contains the full copyright notices and license terms.
-from ..model import ModelSQL, ModelSingleton, fields
-from ..cache import Cache
-from ..config import config
-
-__all__ = ['Configuration']
+from trytond.cache import Cache
+from trytond.config import config
+from trytond.model import ModelSQL, ModelSingleton, fields
 
 
 class Configuration(ModelSingleton, ModelSQL):
@@ -30,6 +28,10 @@ class Configuration(ModelSingleton, ModelSQL):
         cls._get_language_cache.set(None, language)
         return language
 
+    def check(self):
+        "Check configuration coherence on pool initialisation"
+        pass
+
     @classmethod
     def create(cls, vlist):
         records = super().create(vlist)
