@@ -1,38 +1,37 @@
 # This file is part of Tryton.  The COPYRIGHT file at the top level of
 # this repository contains the full copyright notices and license terms.
-import re
 import heapq
 import json
+import logging
+import re
 
-from sql import Null
+from sql import Null, Literal
 from sql.aggregate import Max
 from sql.conditionals import Case
 from collections import defaultdict
 from itertools import groupby
 
-from ..model import (ModelView, ModelSQL, Workflow, fields, Unique,
+from trytond.cache import Cache
+from trytond.i18n import gettext
+from trytond.model import (
+    ModelView, ModelSQL, Workflow, DeactivableMixin, fields, Unique,
     EvalEnvironment)
-from ..report import Report
-from ..wizard import Wizard, StateView, StateAction, Button
-from ..transaction import Transaction
-from ..cache import Cache
-from ..pool import Pool
-from ..pyson import Bool, Eval, PYSONDecoder
-from ..rpc import RPC
-from .. import backend
-from ..protocols.jsonrpc import JSONDecoder, JSONEncoder
-from ..tools import is_instance_method, cursor_dict, grouped_slice
-try:
-    from ..tools.StringMatcher import StringMatcher
-except ImportError:
-    from difflib import SequenceMatcher as StringMatcher
+from trytond.model.exceptions import AccessError, ValidationError
+from trytond.pool import Pool
+from trytond.protocols.jsonrpc import JSONDecoder, JSONEncoder
+from trytond.pyson import Bool, Eval, PYSONDecoder
+from trytond.report import Report
+from trytond.rpc import RPC
+from trytond.tools import is_instance_method, cursor_dict, grouped_slice
+from trytond.tools.string_ import StringMatcher
+from trytond.transaction import Transaction
+from trytond.wizard import Wizard, StateView, StateAction, Button
 
-__all__ = [
-    'Model', 'ModelField', 'ModelAccess', 'ModelFieldAccess', 'ModelButton',
-    'ModelButtonRule', 'ModelButtonClick', 'ModelButtonReset',
-    'ModelData', 'PrintModelGraphStart', 'PrintModelGraph', 'ModelGraph',
-    'ModelWorkflowGraph',
-    ]
+logger = logging.getLogger(__name__)
+
+
+class ConditionError(ValidationError):
+    pass
 
 
 class Model(ModelSQL, ModelView):
@@ -55,10 +54,11 @@ class Model(ModelSQL, ModelView):
             },
         depends=['module'])
     module = fields.Char('Module',
-       help="Module in which this model is defined", readonly=True)
+       help="Module in which this model is defined.", readonly=True)
     global_search_p = fields.Boolean('Global Search')
     fields = fields.One2Many('ir.model.field', 'model', 'Fields',
        required=True)
+    _get_names_cache = Cache('ir.model.get_names')
 
     @classmethod
     def __setup__(cls):
@@ -72,6 +72,7 @@ class Model(ModelSQL, ModelView):
         cls.__rpc__.update({
                 'list_models': RPC(),
                 'list_history': RPC(),
+                'get_names': RPC(),
                 'global_search': RPC(),
                 })
 
@@ -93,7 +94,8 @@ class Model(ModelSQL, ModelView):
             cursor.execute(*ir_model.insert(
                     [ir_model.model, ir_model.name, ir_model.info,
                         ir_model.module],
-                    [[model.__name__, model._get_name(), model.__doc__,
+                    [[
+                            model.__name__, model._get_name(), model.__doc__,
                             module_name]]))
             cursor.execute(*ir_model.select(ir_model.id,
                     where=ir_model.model == model.__name__))
@@ -103,8 +105,29 @@ class Model(ModelSQL, ModelView):
                     [ir_model.name, ir_model.info],
                     [model._get_name(), model.__doc__],
                     where=ir_model.id == model_id))
+        cls._get_names_cache.clear()
         return model_id
 
+    @classmethod
+    def clean(cls):
+        pool = Pool()
+        transaction = Transaction()
+        cursor = transaction.connection.cursor()
+        ir_model = cls.__table__()
+        cursor.execute(*ir_model.select(ir_model.model, ir_model.id))
+        for model, id_ in cursor:
+            try:
+                pool.get(model)
+            except KeyError:
+                logger.info("remove model: %s", model)
+                try:
+                    cls.delete([cls(id_)])
+                    transaction.commit()
+                except Exception:
+                    transaction.rollback()
+                    logger.error(
+                        "could not delete model: %s", model, exc_info=True)
+
     @classmethod
     def list_models(cls):
         'Return a list of all models names'
@@ -121,6 +144,25 @@ class Model(ModelSQL, ModelView):
         return [name for name, model in Pool().iterobject()
             if getattr(model, '_history', False)]
 
+    @classmethod
+    def get_name_items(cls):
+        "Return a list of couple mapping models to names"
+        items = cls._get_names_cache.get('items')
+        if items is None:
+            models = cls.search([])
+            items = [(m.model, m.name) for m in models]
+            cls._get_names_cache.set('items', items)
+        return items
+
+    @classmethod
+    def get_names(cls):
+        "Return a dictionary mapping models to names"
+        dict_ = cls._get_names_cache.get('dict')
+        if dict_ is None:
+            dict_ = dict(cls.get_name_items())
+            cls._get_names_cache.set('dict', dict_)
+        return dict_
+
     @classmethod
     def global_search(cls, text, limit, menu='ir.ui.menu'):
         """
@@ -159,6 +201,10 @@ class Model(ModelSQL, ModelView):
                         record.id, name, icon)
         return heapq.nlargest(int(limit), generate())
 
+    @classmethod
+    def get_name(cls, model):
+        return cls.get_names().get(model, model)
+
 
 class ModelField(ModelSQL, ModelView):
     "Model field"
@@ -190,15 +236,23 @@ class ModelField(ModelSQL, ModelView):
             'readonly': Bool(Eval('module')),
             },
         depends=['module'])
-    groups = fields.Many2Many('ir.model.field-res.group', 'field',
-            'group', 'Groups')
     help = fields.Text('Help', translate=True, loading='lazy',
         states={
             'readonly': Bool(Eval('module')),
             },
         depends=['module'])
     module = fields.Char('Module',
-       help="Module in which this field is defined")
+       help="Module in which this field is defined.")
+    access = fields.Boolean(
+        "Access",
+        states={
+            'invisible': ~Eval('relation'),
+            },
+        depends=['relation'],
+        help="If checked, the access right on the model of the field "
+        "is also tested against the relation of the field.")
+
+    _get_name_cache = Cache('ir.model.field.get_name')
 
     @classmethod
     def __setup__(cls):
@@ -228,37 +282,69 @@ class ModelField(ModelSQL, ModelView):
                 ir_model_field.relation.as_('relation'),
                 ir_model_field.module.as_('module'),
                 ir_model_field.help.as_('help'),
+                ir_model_field.access.as_('access'),
                 where=ir_model.model == model.__name__))
         model_fields = {f['name']: f for f in cursor_dict(cursor)}
 
-        for field_name, field in model._fields.iteritems():
+        for field_name, field in model._fields.items():
             if hasattr(field, 'model_name'):
                 relation = field.model_name
             elif hasattr(field, 'relation_name'):
                 relation = field.relation_name
             else:
                 relation = None
+            access = field_name in model.__access__
 
             if field_name not in model_fields:
                 cursor.execute(*ir_model_field.insert(
                         [ir_model_field.model, ir_model_field.name,
                             ir_model_field.field_description,
                             ir_model_field.ttype, ir_model_field.relation,
-                            ir_model_field.help, ir_model_field.module],
-                        [[model_id, field_name, field.string, field._type,
-                                relation, field.help, module_name]]))
+                            ir_model_field.help, ir_model_field.module,
+                            ir_model_field.access],
+                        [[
+                                model_id, field_name,
+                                field.string,
+                                field._type, relation,
+                                field.help, module_name,
+                                access]]))
             elif (model_fields[field_name]['field_description'] != field.string
                     or model_fields[field_name]['ttype'] != field._type
                     or model_fields[field_name]['relation'] != relation
-                    or model_fields[field_name]['help'] != field.help):
+                    or model_fields[field_name]['help'] != field.help
+                    or model_fields[field_name]['access'] != access):
                 cursor.execute(*ir_model_field.update(
                         [ir_model_field.field_description,
                             ir_model_field.ttype, ir_model_field.relation,
-                            ir_model_field.help],
-                        [field.string, field._type, relation, field.help],
-                        where=ir_model_field.id ==
-                        model_fields[field_name]['id']))
+                            ir_model_field.help, ir_model_field.access],
+                        [field.string, field._type, relation, field.help,
+                            access],
+                        where=(ir_model_field.id
+                            == model_fields[field_name]['id'])))
 
+    @classmethod
+    def clean(cls):
+        pool = Pool()
+        IrModel = pool.get('ir.model')
+        transaction = Transaction()
+        cursor = transaction.connection.cursor()
+        ir_model = IrModel.__table__()
+        ir_model_field = cls.__table__()
+        cursor.execute(*ir_model_field
+            .join(ir_model, condition=ir_model_field.model == ir_model.id)
+            .select(ir_model.model, ir_model_field.name, ir_model_field.id))
+        for model, field, id_ in cursor:
+            Model = pool.get(model)
+            if field not in Model._fields:
+                logger.info("remove field: %s.%s", model, field)
+                try:
+                    cls.delete([cls(id_)])
+                    transaction.commit()
+                except Exception:
+                    transaction.rollback()
+                    logger.error(
+                        "could not delete field: %s.%s", model, field,
+                        exc_info=True)
 
     @staticmethod
     def default_name():
@@ -268,17 +354,47 @@ class ModelField(ModelSQL, ModelView):
     def default_field_description():
         return 'No description available'
 
+    def get_rec_name(self, name):
+        if self.field_description:
+            return '%s (%s)' % (self.field_description, self.name)
+        else:
+            return self.name
+
     @classmethod
-    def read(cls, ids, fields_names=None):
+    def search_rec_name(cls, name, clause):
+        if clause[1].startswith('!') or clause[1].startswith('not '):
+            bool_op = 'AND'
+        else:
+            bool_op = 'OR'
+        return [bool_op,
+            ('field_description',) + tuple(clause[1:]),
+            ('name',) + tuple(clause[1:]),
+            ]
+
+    @classmethod
+    def get_name(cls, model, field):
+        name = cls._get_name_cache.get((model, field))
+        if name is None:
+            fields = cls.search([
+                    ('model.model', '=', model),
+                    ('name', '=', field),
+                    ], limit=1)
+            if fields:
+                field, = fields
+                name = field.field_description
+                cls._get_name_cache.set((model, field), name)
+            else:
+                name = field
+        return name
+
+    @classmethod
+    def read(cls, ids, fields_names):
         pool = Pool()
         Translation = pool.get('ir.translation')
         Model = pool.get('ir.model')
 
         to_delete = []
         if Transaction().context.get('language'):
-            if fields_names is None:
-                fields_names = cls._fields.keys()
-
             if 'field_description' in fields_names \
                     or 'help' in fields_names:
                 if 'model' not in fields_names:
@@ -288,7 +404,7 @@ class ModelField(ModelSQL, ModelView):
                     fields_names.append('name')
                     to_delete.append('name')
 
-        res = super(ModelField, cls).read(ids, fields_names=fields_names)
+        res = super(ModelField, cls).read(ids, fields_names)
 
         if (Transaction().context.get('language')
                 and ('field_description' in fields_names
@@ -304,7 +420,7 @@ class ModelField(ModelSQL, ModelView):
             model = Model.__table__()
             cursor.execute(*model.select(model.id, model.model,
                     where=model.id.in_(model_ids)))
-            id2model = dict(cursor.fetchall())
+            id2model = dict(cursor)
 
             trans_args = []
             for rec in res:
@@ -344,7 +460,7 @@ class ModelField(ModelSQL, ModelView):
         return res
 
 
-class ModelAccess(ModelSQL, ModelView):
+class ModelAccess(DeactivableMixin, ModelSQL, ModelView):
     "Model access"
     __name__ = 'ir.model.access'
     model = fields.Many2One('ir.model', 'Model', required=True,
@@ -361,27 +477,10 @@ class ModelAccess(ModelSQL, ModelView):
     @classmethod
     def __setup__(cls):
         super(ModelAccess, cls).__setup__()
-        cls._error_messages.update({
-            'read': 'You can not read this document! (%s)',
-            'write': 'You can not write in this document! (%s)',
-            'create': 'You can not create this kind of document! (%s)',
-            'delete': 'You can not delete this document! (%s)',
-            })
         cls.__rpc__.update({
                 'get_access': RPC(),
                 })
 
-    @classmethod
-    def __register__(cls, module_name):
-        TableHandler = backend.get('TableHandler')
-
-        super(ModelAccess, cls).__register__(module_name)
-
-        table = TableHandler(cls, module_name)
-
-        # Migration from 2.6 (model, group) no more unique
-        table.drop_constraint('model_group_uniq')
-
     @staticmethod
     def check_xml_record(accesses, values):
         return True
@@ -419,11 +518,13 @@ class ModelAccess(ModelSQL, ModelView):
         pool = Pool()
         Model = pool.get('ir.model')
         UserGroup = pool.get('res.user-res.group')
+        Group = pool.get('res.group')
         cursor = Transaction().connection.cursor()
         user = Transaction().user
         model_access = cls.__table__()
         ir_model = Model.__table__()
-        user_group = UserGroup.__table__()
+        user_group = UserGroup.user_group_all_table()
+        group = Group.__table__()
 
         access = {}
         for model in models:
@@ -434,31 +535,67 @@ class ModelAccess(ModelSQL, ModelView):
         else:
             return access
 
+        def fill_models(Model, models):
+            if Model.__name__ in models:
+                return
+            models.append(Model.__name__)
+            for field_name in Model.__access__:
+                field = getattr(Model, field_name)
+                fill_models(field.get_target(), models)
+        model2models = defaultdict(list)
+        for model in models:
+            fill_models(pool.get(model), model2models[model])
+
+        all_models = list(set(sum(model2models.values(), [])))
         default = {'read': True, 'write': True, 'create': True, 'delete': True}
         access = dict((m, default) for m in models)
         cursor.execute(*model_access.join(ir_model, 'LEFT',
                 condition=model_access.model == ir_model.id
                 ).join(user_group, 'LEFT',
                 condition=user_group.group == model_access.group
+                ).join(group, 'LEFT',
+                condition=model_access.group == group.id
                 ).select(
                 ir_model.model,
-                Max(Case((model_access.perm_read == True, 1), else_=0)),
-                Max(Case((model_access.perm_write == True, 1), else_=0)),
-                Max(Case((model_access.perm_create == True, 1), else_=0)),
-                Max(Case((model_access.perm_delete == True, 1), else_=0)),
-                where=ir_model.model.in_(models)
-                & ((user_group.user == user) | (model_access.group == Null)),
+                Max(Case(
+                        (model_access.perm_read == Literal(True), 1),
+                        else_=0)),
+                Max(Case(
+                        (model_access.perm_write == Literal(True), 1),
+                        else_=0)),
+                Max(Case(
+                        (model_access.perm_create == Literal(True), 1),
+                        else_=0)),
+                Max(Case(
+                        (model_access.perm_delete == Literal(True), 1),
+                        else_=0)),
+                where=ir_model.model.in_(all_models)
+                & (model_access.active == Literal(True))
+                & ((
+                        (user_group.user == user)
+                        & (group.active == Literal(True)))
+                    | (model_access.group == Null)),
                 group_by=ir_model.model))
-        access.update(dict(
-                (m, {'read': r, 'write': w, 'create': c, 'delete': d})
-                for m, r, w, c, d in cursor.fetchall()))
-        for model, maccess in access.iteritems():
+        raw_access = {
+            m: {'read': r, 'write': w, 'create': c, 'delete': d}
+            for m, r, w, c, d in cursor}
+
+        for model in models:
+            access[model] = {
+                perm: max(
+                    (raw_access[m][perm] for m in model2models[model]
+                        if m in raw_access),
+                    default=True)
+                for perm in ['read', 'write', 'create', 'delete']}
+        for model, maccess in access.items():
             cls._get_access_cache.set((user, model), maccess)
         return access
 
     @classmethod
     def check(cls, model_name, mode='read', raise_exception=True):
         'Check access for model_name and mode'
+        pool = Pool()
+        Model = pool.get(model_name)
         assert mode in ['read', 'write', 'create', 'delete'], \
             'Invalid access mode for security'
         if ((Transaction().user == 0)
@@ -469,7 +606,8 @@ class ModelAccess(ModelSQL, ModelView):
         access = cls.get_access([model_name])[model_name][mode]
         if not access and access is not None:
             if raise_exception:
-                cls.raise_user_error(mode, model_name)
+                raise AccessError(gettext(
+                        'ir.msg_access_rule_error', **Model.__names__()))
             else:
                 return False
         return True
@@ -496,7 +634,7 @@ class ModelAccess(ModelSQL, ModelView):
                 return True
         elif field._type == 'reference':
             selection = field.selection
-            if isinstance(selection, basestring):
+            if isinstance(selection, str):
                 sel_func = getattr(Model, field.selection)
                 if not is_instance_method(Model, field.selection):
                     selection = sel_func()
@@ -504,7 +642,7 @@ class ModelAccess(ModelSQL, ModelView):
                     # XXX Can not check access right on instance method
                     selection = []
             for model_name, _ in selection:
-                if not cls.check(model_name, mode=mode,
+                if model_name and not cls.check(model_name, mode=mode,
                         raise_exception=False):
                     return False
             return True
@@ -534,7 +672,7 @@ class ModelAccess(ModelSQL, ModelView):
         ModelView._fields_view_get_cache.clear()
 
 
-class ModelFieldAccess(ModelSQL, ModelView):
+class ModelFieldAccess(DeactivableMixin, ModelSQL, ModelView):
     "Model Field Access"
     __name__ = 'ir.model.field.access'
     field = fields.Many2One('ir.model.field', 'Field', required=True,
@@ -547,25 +685,6 @@ class ModelFieldAccess(ModelSQL, ModelView):
     description = fields.Text('Description')
     _get_access_cache = Cache('ir_model_field_access.check', context=False)
 
-    @classmethod
-    def __setup__(cls):
-        super(ModelFieldAccess, cls).__setup__()
-        cls._error_messages.update({
-            'read': 'You can not read the field! (%s.%s)',
-            'write': 'You can not write on the field! (%s.%s)',
-            })
-
-    @classmethod
-    def __register__(cls, module_name):
-        TableHandler = backend.get('TableHandler')
-
-        super(ModelFieldAccess, cls).__register__(module_name)
-
-        table = TableHandler(cls, module_name)
-
-        # Migration from 2.6 (field, group) no more unique
-        table.drop_constraint('field_group_uniq')
-
     @staticmethod
     def check_xml_record(field_accesses, values):
         return True
@@ -605,11 +724,13 @@ class ModelFieldAccess(ModelSQL, ModelView):
         Model = pool.get('ir.model')
         ModelField = pool.get('ir.model.field')
         UserGroup = pool.get('res.user-res.group')
+        Group = pool.get('res.group')
         user = Transaction().user
         field_access = cls.__table__()
         ir_model = Model.__table__()
         model_field = ModelField.__table__()
-        user_group = UserGroup.__table__()
+        user_group = UserGroup.user_group_all_table()
+        group = Group.__table__()
 
         accesses = {}
         for model in models:
@@ -629,19 +750,33 @@ class ModelFieldAccess(ModelSQL, ModelView):
                 condition=model_field.model == ir_model.id
                 ).join(user_group, 'LEFT',
                 condition=user_group.group == field_access.group
+                ).join(group, 'LEFT',
+                condition=field_access.group == group.id
                 ).select(
                 ir_model.model,
                 model_field.name,
-                Max(Case((field_access.perm_read == True , 1), else_=0)),
-                Max(Case((field_access.perm_write == True, 1), else_=0)),
-                Max(Case((field_access.perm_create == True, 1), else_=0)),
-                Max(Case((field_access.perm_delete == True, 1), else_=0)),
+                Max(Case(
+                        (field_access.perm_read == Literal(True), 1),
+                        else_=0)),
+                Max(Case(
+                        (field_access.perm_write == Literal(True), 1),
+                        else_=0)),
+                Max(Case(
+                        (field_access.perm_create == Literal(True), 1),
+                        else_=0)),
+                Max(Case(
+                        (field_access.perm_delete == Literal(True), 1),
+                        else_=0)),
                 where=ir_model.model.in_(models)
-                & ((user_group.user == user) | (field_access.group == Null)),
+                & (field_access.active == Literal(True))
+                & ((
+                        (user_group.user == user)
+                        & (group.active == Literal(True)))
+                    | (field_access.group == Null)),
                 group_by=[ir_model.model, model_field.name]))
-        for m, f, r, w, c, d in cursor.fetchall():
+        for m, f, r, w, c, d in cursor:
             accesses[m][f] = {'read': r, 'write': w, 'create': c, 'delete': d}
-        for model, maccesses in accesses.iteritems():
+        for model, maccesses in accesses.items():
             cls._get_access_cache.set((user, model), maccesses)
         return accesses
 
@@ -651,6 +786,8 @@ class ModelFieldAccess(ModelSQL, ModelView):
         '''
         Check access for fields on model_name.
         '''
+        pool = Pool()
+        Model = pool.get(model_name)
         assert mode in ('read', 'write', 'create', 'delete'), \
             'Invalid access mode'
         if ((Transaction().user == 0)
@@ -661,13 +798,15 @@ class ModelFieldAccess(ModelSQL, ModelView):
             return True
 
         accesses = dict((f, a[mode])
-            for f, a in cls.get_access([model_name])[model_name].iteritems())
+            for f, a in cls.get_access([model_name])[model_name].items())
         if access:
             return accesses
         for field in fields:
             if not accesses.get(field, True):
                 if raise_exception:
-                    cls.raise_user_error(mode, (model_name, field))
+                    raise AccessError(
+                        gettext('ir.msg_access_rule_field_error',
+                            **Model.__names__(field)))
                 else:
                     return False
         return True
@@ -699,6 +838,10 @@ class ModelButton(ModelSQL, ModelView):
     "Model Button"
     __name__ = 'ir.model.button'
     name = fields.Char('Name', required=True, readonly=True)
+    string = fields.Char("Label", translate=True)
+    help = fields.Text("Help", translate=True)
+    confirm = fields.Text("Confirm", translate=True,
+        help="Text to ask user confirmation when clicking the button.")
     model = fields.Many2One('ir.model', 'Model', required=True, readonly=True,
         ondelete='CASCADE', select=True)
     groups = fields.Many2Many('ir.model.button-res.group', 'button', 'group',
@@ -714,7 +857,7 @@ class ModelButton(ModelSQL, ModelView):
             ('id', '!=', Eval('id', -1)),
             ],
         depends=['model', 'id'],
-        help="Button that should reset the rules")
+        help="Button that should reset the rules.")
     reset = fields.Many2Many(
         'ir.model.button-button.reset', 'button', 'button_ruled', "Reset",
         domain=[
@@ -723,6 +866,8 @@ class ModelButton(ModelSQL, ModelView):
             ],
         depends=['model', 'id'])
     _reset_cache = Cache('ir.model.button.reset')
+    _view_attributes_cache = Cache(
+        'ir.model.button.view_attributes', context=False)
 
     @classmethod
     def __setup__(cls):
@@ -741,6 +886,7 @@ class ModelButton(ModelSQL, ModelView):
         cls._groups_cache.clear()
         cls._rules_cache.clear()
         cls._reset_cache.clear()
+        cls._view_attributes_cache.clear()
         return result
 
     @classmethod
@@ -750,6 +896,7 @@ class ModelButton(ModelSQL, ModelView):
         cls._groups_cache.clear()
         cls._rules_cache.clear()
         cls._reset_cache.clear()
+        cls._view_attributes_cache.clear()
 
     @classmethod
     def delete(cls, buttons):
@@ -758,6 +905,7 @@ class ModelButton(ModelSQL, ModelView):
         cls._groups_cache.clear()
         cls._rules_cache.clear()
         cls._reset_cache.clear()
+        cls._view_attributes_cache.clear()
 
     @classmethod
     def copy(cls, buttons, default=None):
@@ -829,6 +977,29 @@ class ModelButton(ModelSQL, ModelView):
         cls._reset_cache.set(key, reset)
         return reset
 
+    @classmethod
+    def get_view_attributes(cls, model, name):
+        "Return the view attributes of the named button of the model"
+        key = (model, name, Transaction().language)
+        attributes = cls._view_attributes_cache.get(key)
+        if attributes is not None:
+            return attributes
+        buttons = cls.search([
+                ('model.model', '=', model),
+                ('name', '=', name),
+                ])
+        if not buttons:
+            attributes = {}
+        else:
+            button, = buttons
+            attributes = {
+                'string': button.string,
+                'help': button.help,
+                'confirm': button.confirm,
+                }
+        cls._view_attributes_cache.set(key, attributes)
+        return attributes
+
 
 class ModelButtonRule(ModelSQL, ModelView):
     "Model Button Rule"
@@ -842,14 +1013,6 @@ class ModelButtonRule(ModelSQL, ModelView):
         help='A PYSON statement evaluated with the record represented by '
         '"self"\nIt activate the rule if true.')
 
-    @classmethod
-    def __setup__(cls):
-        super(ModelButtonRule, cls).__setup__()
-        cls._error_messages.update({
-                'invalid_condition': ('Condition "%(condition)s" is not a '
-                    'valid PYSON expression on button rule "%(rule)s".'),
-                })
-
     @classmethod
     def default_number_user(cls):
         return 1
@@ -867,10 +1030,10 @@ class ModelButtonRule(ModelSQL, ModelView):
             try:
                 PYSONDecoder(noeval=True).decode(rule.condition)
             except Exception:
-                cls.raise_user_error('invalid_condition', {
-                        'condition': rule.condition,
-                        'rule': rule.rec_name,
-                        })
+                raise ConditionError(
+                    gettext('ir.msg_model_invalid_condition',
+                        condition=rule.condition,
+                        rule=rule.rec_name))
 
     def test(self, record, clicks):
         "Test if the rule passes for the record"
@@ -911,13 +1074,12 @@ class ModelButtonRule(ModelSQL, ModelView):
         ModelButton._rules_cache.clear()
 
 
-class ModelButtonClick(ModelSQL, ModelView):
+class ModelButtonClick(DeactivableMixin, ModelSQL, ModelView):
     "Model Button Click"
     __name__ = 'ir.model.button.click'
     button = fields.Many2One(
         'ir.model.button', "Button", required=True, ondelete='CASCADE')
     record_id = fields.Integer("Record ID", required=True)
-    active = fields.Boolean("Active")
 
     @classmethod
     def __setup__(cls):
@@ -926,10 +1088,6 @@ class ModelButtonClick(ModelSQL, ModelView):
                 'get_click': RPC(),
                 })
 
-    @classmethod
-    def default_active(cls):
-        return True
-
     @classmethod
     def register(cls, model, name, records):
         pool = Pool()
@@ -1003,15 +1161,19 @@ class ModelData(ModelSQL, ModelView):
         select=True)
     model = fields.Char('Model', required=True, select=True)
     module = fields.Char('Module', required=True, select=True)
-    db_id = fields.Integer('Resource ID',
-        help="The id of the record in the database.", select=True,
-        required=True)
+    db_id = fields.Integer(
+        "Resource ID", select=True,
+        states={
+            'required': ~Eval('noupdate', False),
+            },
+        help="The id of the record in the database.")
     values = fields.Text('Values')
     fs_values = fields.Text('Values on File System')
     noupdate = fields.Boolean('No Update')
     out_of_sync = fields.Function(fields.Boolean('Out of Sync'),
         'get_out_of_sync', searcher='search_out_of_sync')
     _get_id_cache = Cache('ir_model_data.get_id', context=False)
+    _has_model_cache = Cache('ir_model_data.has_model', context=False)
 
     @classmethod
     def __setup__(cls):
@@ -1025,24 +1187,31 @@ class ModelData(ModelSQL, ModelView):
         cls._buttons.update({
                 'sync': {
                     'invisible': ~Eval('out_of_sync'),
+                    'depends': ['out_of_sync'],
                     },
                 })
+        cls.__rpc__.update({
+                'sync': RPC(
+                    readonly=False, instantiate=0, fresh_session=True),
+                })
 
     @classmethod
     def __register__(cls, module_name):
-        TableHandler = backend.get('TableHandler')
         cursor = Transaction().connection.cursor()
         model_data = cls.__table__()
 
         super(ModelData, cls).__register__(module_name)
 
-        table = TableHandler(cls, module_name)
+        table_h = cls.__table_handler__(module_name)
 
-        # Migration from 2.6: remove inherit
-        if table.column_exist('inherit'):
-            cursor.execute(*model_data.delete(
-                    where=model_data.inherit == True))
-            table.drop_column('inherit', True)
+        # Migration from 4.6: register buttons on ir module
+        cursor.execute(*model_data.update(
+                [model_data.module], ['ir'],
+                where=((model_data.module == 'res')
+                    & (model_data.fs_id == 'model_data_sync_button'))))
+
+        # Migration from 5.0: remove required on db_id
+        table_h.not_null_action('db_id', action='remove')
 
     @staticmethod
     def default_noupdate():
@@ -1062,17 +1231,43 @@ class ModelData(ModelSQL, ModelView):
                 value))
         return [('id', 'in', query)]
 
+    @classmethod
+    def create(cls, *args):
+        records = super(ModelData, cls).create(*args)
+        cls._has_model_cache.clear()
+        return records
+
     @classmethod
     def write(cls, data, values, *args):
         super(ModelData, cls).write(data, values, *args)
         # Restart the cache for get_id
         cls._get_id_cache.clear()
+        cls._has_model_cache.clear()
 
     @classmethod
-    def get_id(cls, module, fs_id):
+    def delete(cls, records):
+        super(ModelData, cls).delete(records)
+        cls._has_model_cache.clear()
+
+    @classmethod
+    def has_model(cls, model):
+        models = cls._has_model_cache.get(None)
+        if models is None:
+            table = cls.__table__()
+            cursor = Transaction().connection.cursor()
+
+            cursor.execute(*table.select(table.model, group_by=[table.model]))
+            models = [m for m, in cursor]
+            cls._has_model_cache.set(None, models)
+        return model in models
+
+    @classmethod
+    def get_id(cls, module, fs_id=None):
         """
         Return for an fs_id the corresponding db_id.
         """
+        if fs_id is None:
+            module, fs_id = module.split('.', 1)
         key = (module, fs_id)
         id_ = cls._get_id_cache.get(key)
         if id_ is not None:
@@ -1091,7 +1286,8 @@ class ModelData(ModelSQL, ModelView):
     @classmethod
     def dump_values(cls, values):
         return json.dumps(
-            sorted(values.iteritems()), cls=JSONEncoder, separators=(',', ':'))
+            sorted(values.items()), cls=JSONEncoder, separators=(',', ':'),
+            sort_keys=True)
 
     @classmethod
     def load_values(cls, values):
@@ -1109,25 +1305,43 @@ class ModelData(ModelSQL, ModelView):
     @classmethod
     @ModelView.button
     def sync(cls, records):
-        pool = Pool()
-        to_write = []
-        models_to_write = defaultdict(list)
-        for data in records:
-            Model = pool.get(data.model)
-            values = cls.load_values(data.values)
-            fs_values = cls.load_values(data.fs_values)
-            # values could be the same once loaded
-            # if they come from version < 3.2
-            if values != fs_values:
-                record = Model(data.db_id)
-                models_to_write[Model].extend(([record], fs_values))
-            to_write.extend([[data], {
-                        'values': cls.dump_values(fs_values),
-                        }])
-        for Model, values_to_write in models_to_write.iteritems():
-            Model.write(*values_to_write)
-        if to_write:
-            cls.write(*to_write)
+        def settable(Model, fieldname):
+            try:
+                field = Model._fields[fieldname]
+            except KeyError:
+                return False
+
+            if isinstance(field, fields.Function) and not field.setter:
+                return False
+
+            return True
+
+        with Transaction().set_user(0):
+            pool = Pool()
+            to_write = []
+            models_to_write = defaultdict(list)
+            for data in records:
+                try:
+                    Model = pool.get(data.model)
+                except KeyError:
+                    continue
+                values = cls.load_values(data.values)
+                fs_values = cls.load_values(data.fs_values)
+                # values could be the same once loaded
+                # if they come from version < 3.2
+                if values != fs_values:
+                    values = {f: v for f, v in fs_values.items()
+                        if settable(Model, f)}
+                    record = Model(data.db_id)
+                    models_to_write[Model].extend(([record], values))
+                to_write.extend([[data], {
+                            'values': cls.dump_values(fs_values),
+                            'fs_values': cls.dump_values(fs_values),
+                            }])
+            for Model, values_to_write in models_to_write.items():
+                Model.write(*values_to_write)
+            if to_write:
+                cls.write(*to_write)
 
 
 class PrintModelGraphStart(ModelView):
@@ -1222,7 +1436,7 @@ class ModelGraph(Report):
 
         for model in models:
             if filter and re.search(filter, model.model):
-                    continue
+                continue
             label = '"{' + model.model + '\\n'
             if model.fields:
                 label += '|'
@@ -1233,7 +1447,7 @@ class ModelGraph(Report):
                 label += '+ ' + field.name + ': ' + field.ttype
                 if field.relation:
                     label += ' ' + field.relation
-                label += '\l'
+                label += '\\l'
             label += '}"'
             node_name = '"%s"' % model.model
             node = pydot.Node(node_name, shape='record', label=label)
