@@ -5,7 +5,11 @@ from functools import wraps
 
 from sql.operators import NotIn
 
-from trytond.model import ModelView, ModelSQL, fields, Unique
+from trytond.cache import Cache
+from trytond.exceptions import UserError
+from trytond.i18n import gettext
+from trytond.model import ModelView, ModelSQL, fields, Unique, sequence_ordered
+from trytond.model.exceptions import AccessError
 from trytond.modules import get_module_list, get_module_info
 from trytond.wizard import Wizard, StateView, Button, StateTransition, \
     StateAction
@@ -15,13 +19,9 @@ from trytond.transaction import Transaction
 from trytond.pyson import Eval
 from trytond.rpc import RPC
 
-__all__ = [
-    'Module', 'ModuleDependency', 'ModuleConfigWizardItem',
-    'ModuleConfigWizardFirst', 'ModuleConfigWizardOther',
-    'ModuleConfigWizardDone', 'ModuleConfigWizard',
-    'ModuleActivateUpgradeStart', 'ModuleActivateUpgradeDone',
-    'ModuleActivateUpgrade', 'ModuleConfig',
-    ]
+
+class DeactivateDependencyError(UserError):
+    pass
 
 
 def filter_state(state):
@@ -65,12 +65,6 @@ class Module(ModelSQL, ModelView):
         cls.__rpc__.update({
                 'on_write': RPC(instantiate=0),
                 })
-        cls._error_messages.update({
-            'delete_state': ('You can not remove a module that is activated '
-                    'or will be activated'),
-            'deactivate_dep': ('Some activated modules depend on the ones '
-                    'you are trying to deactivate:'),
-            })
         cls._buttons.update({
                 'activate': {
                     'invisible': Eval('state') != 'not activated',
@@ -102,15 +96,14 @@ class Module(ModelSQL, ModelView):
     def __register__(cls, module_name):
         pool = Pool()
         ModelData = pool.get('ir.model.data')
-        TableHandler = backend.get('TableHandler')
         sql_table = cls.__table__()
         model_data_sql_table = ModelData.__table__()
         cursor = Transaction().connection.cursor()
 
         # Migration from 3.6: remove double module
         old_table = 'ir_module_module'
-        if TableHandler.table_exist(old_table):
-            TableHandler.table_rename(old_table, cls._table)
+        if backend.TableHandler.table_exist(old_table):
+            backend.TableHandler.table_rename(old_table, cls._table)
 
         super(Module, cls).__register__(module_name)
 
@@ -176,7 +169,7 @@ class Module(ModelSQL, ModelView):
                     'to remove',
                     'to activate',
                     ):
-                cls.raise_user_error('delete_state')
+                raise AccessError(gettext('ir.msg_module_delete_state'))
         return super(Module, cls).delete(records)
 
     @classmethod
@@ -265,9 +258,9 @@ class Module(ModelSQL, ModelView):
                         module_table.state, ['not activated', 'to remove'])))
             res = cursor.fetchall()
             if res:
-                cls.raise_user_error('deactivate_dep',
-                        error_description='\n'.join(
-                            '\t%s: %s' % (x[0], x[1]) for x in res))
+                raise DeactivateDependencyError(
+                    gettext('ir.msg_module_deactivate_dependency'),
+                    '\n'.join('\t%s: %s' % (x[0], x[1]) for x in res))
         cls.write(modules, {'state': 'to remove'})
 
     @classmethod
@@ -360,12 +353,10 @@ class ModuleDependency(ModelSQL, ModelView):
 
     @classmethod
     def __register__(cls, module_name):
-        TableHandler = backend.get('TableHandler')
-
         # Migration from 3.6: remove double module
         old_table = 'ir_module_module_dependency'
-        if TableHandler.table_exist(old_table):
-            TableHandler.table_rename(old_table, cls._table)
+        if backend.TableHandler.table_exist(old_table):
+            backend.TableHandler.table_rename(old_table, cls._table)
 
         super(ModuleDependency, cls).__register__(module_name)
 
@@ -381,25 +372,18 @@ class ModuleDependency(ModelSQL, ModelView):
             return 'unknown'
 
 
-class ModuleConfigWizardItem(ModelSQL, ModelView):
+class ModuleConfigWizardItem(sequence_ordered(), ModelSQL, ModelView):
     "Config wizard to run after activating a module"
     __name__ = 'ir.module.config_wizard.item'
     action = fields.Many2One('ir.action', 'Action', required=True,
         readonly=True)
-    sequence = fields.Integer('Sequence', required=True)
     state = fields.Selection([
         ('open', 'Open'),
         ('done', 'Done'),
         ], string='State', required=True, select=True)
 
-    @classmethod
-    def __setup__(cls):
-        super(ModuleConfigWizardItem, cls).__setup__()
-        cls._order.insert(0, ('sequence', 'ASC'))
-
     @classmethod
     def __register__(cls, module_name):
-        TableHandler = backend.get('TableHandler')
         cursor = Transaction().connection.cursor()
         pool = Pool()
         ModelData = pool.get('ir.model.data')
@@ -407,16 +391,21 @@ class ModuleConfigWizardItem(ModelSQL, ModelView):
 
         # Migration from 3.6: remove double module
         old_table = 'ir_module_module_config_wizard_item'
-        if TableHandler.table_exist(old_table):
-            TableHandler.table_rename(old_table, cls._table)
+        if backend.TableHandler.table_exist(old_table):
+            backend.TableHandler.table_rename(old_table, cls._table)
         cursor.execute(*model_data.update(
                 columns=[model_data.model],
                 values=[cls.__name__],
-                where=(model_data.model ==
-                    'ir.module.module.config_wizard.item')))
+                where=(model_data.model
+                    == 'ir.module.module.config_wizard.item')))
 
         super(ModuleConfigWizardItem, cls).__register__(module_name)
 
+        table = cls.__table_handler__(module_name)
+
+        # Migration from 5.0: remove required on sequence
+        table.not_null_action('sequence', 'remove')
+
     @staticmethod
     def default_state():
         return 'open'
@@ -569,7 +558,8 @@ class ModuleActivateUpgrade(Wizard):
         pool = Pool()
         Module = pool.get('ir.module')
         Lang = pool.get('ir.lang')
-        with Transaction().new_transaction():
+        transaction = Transaction()
+        with transaction.new_transaction():
             modules = Module.search([
                 ('state', 'in', ['to upgrade', 'to remove', 'to activate']),
                 ])
@@ -580,6 +570,7 @@ class ModuleActivateUpgrade(Wizard):
             lang = [x.code for x in langs]
         if update:
             pool.init(update=update, lang=lang)
+            Cache.refresh_pool(transaction)
         return 'done'
 
 
