@@ -4,6 +4,7 @@
 import inspect
 import copy
 
+from trytond.i18n import gettext
 from trytond.model.fields.field import Field
 from trytond.tools import is_instance_method
 from trytond.transaction import Transaction
@@ -76,7 +77,9 @@ class Function(Field):
             return method(domain, tables)
         if self.searcher:
             return getattr(Model, self.searcher)(self.name, domain)
-        Model.raise_user_error('search_function_missing', self.name)
+        raise NotImplementedError(gettext(
+                'ir.msg_search_function_missing',
+                **Model.__names__(self.name)))
 
     def get(self, ids, Model, name, values=None):
         '''
@@ -87,8 +90,7 @@ class Function(Field):
         with Transaction().set_context(_check_access=False):
             method = getattr(Model, self.getter)
             instance_method = is_instance_method(Model, self.getter)
-            signature = inspect.signature(method)
-            uses_names = 'names' in signature.parameters
+            multiple = self.getter_multiple(method)
 
             def call(name):
                 records = Model.browse(ids)
@@ -98,11 +100,11 @@ class Function(Field):
                     return dict((r.id, method(r, name)) for r in records)
             if isinstance(name, list):
                 names = name
-                if uses_names:
+                if multiple:
                     return call(names)
                 return dict((name, call(name)) for name in names)
             else:
-                if uses_names:
+                if multiple:
                     name = [name]
                 return call(name)
 
@@ -117,10 +119,40 @@ class Function(Field):
                 args = iter((ids, value) + args)
                 for ids, value in zip(args, args):
                     setter(Model.browse(ids), name, value)
+            else:
+                raise NotImplementedError(gettext(
+                        'ir.msg_setter_function_missing',
+                        **Model.__names__(self.name)))
+
+    def __get__(self, inst, cls):
+        try:
+            return super().__get__(inst, cls)
+        except AttributeError:
+            if not self.getter.startswith('on_change_with'):
+                raise
+            value = getattr(inst, self.getter)(self.name)
+            # Use temporary instance to not modify instance values
+            temp_inst = cls()
+            # Set the value to have proper type
+            self.__set__(temp_inst, value)
+            return super().__get__(temp_inst, cls)
 
     def __set__(self, inst, value):
         self._field.__set__(inst, value)
 
+    def definition(self, model, language):
+        definition = self._field.definition(model, language)
+        definition.update(super().definition(model, language))
+        definition['searchable'] &= (
+            bool(self.searcher) or hasattr(model, 'domain_' + self.name))
+        definition['sortable'] &= hasattr(model, 'order_' + self.name)
+        return definition
+
+    def getter_multiple(self, method):
+        "Returns True if getter function accepts multiple fields"
+        signature = inspect.signature(method)
+        return 'names' in signature.parameters
+
 
 class MultiValue(Function):
 
