@@ -9,6 +9,12 @@ class ModelSingleton(ModelStorage):
     Define a singleton model in Tryton.
     """
 
+    @classmethod
+    def __setup__(cls):
+        super().__setup__()
+        # Cache disable because it is used as a read by the client
+        cls.__rpc__['default_get'].cache = None
+
     @classmethod
     def get_singleton(cls):
         '''
@@ -28,24 +34,22 @@ class ModelSingleton(ModelStorage):
         return [singleton]
 
     @classmethod
-    def read(cls, ids, fields_names=None):
+    def read(cls, ids, fields_names):
         singleton = cls.get_singleton()
         if not singleton:
-            if not fields_names:
-                fields_names = list(cls._fields.keys())
-            fname_no_rec_name = [f for f in fields_names if '.' not in f]
+            fname_no_rec_name = [
+                f for f in fields_names
+                if '.' not in f and not f.startswith('_')]
             res = cls.default_get(fname_no_rec_name,
                 with_rec_name=len(fname_no_rec_name) != len(fields_names))
             for field_name in fields_names:
                 if field_name not in res:
                     res[field_name] = None
-            for field_name in list(res.keys()):
-                if field_name not in fields_names:
-                    del res[field_name]
             res['id'] = ids[0]
+            res['_write'] = True
+            res['_delete'] = True
             return [res]
-        res = super(ModelSingleton, cls).read([singleton.id],
-            fields_names=fields_names)
+        res = super(ModelSingleton, cls).read([singleton.id], fields_names)
         res[0]['id'] = ids[0]
         return res
 
@@ -61,9 +65,7 @@ class ModelSingleton(ModelStorage):
         super(ModelSingleton, cls).write(*args)
         # Clean local cache of original records
         for record in sum(actions[0:None:2], []):
-            local_cache = record._local_cache.get(record.id)
-            if local_cache:
-                local_cache.clear()
+            record._local_cache.pop(record.id, None)
         # Clean transaction cache of all ids
         for cache in Transaction().cache.values():
             if cls.__name__ in cache:
@@ -110,6 +112,5 @@ class ModelSingleton(ModelStorage):
                     if cls._fields[field]._type in ('many2one',):
                         fields_names.append(field + '.rec_name')
             default, = cls.read([singleton.id], fields_names=fields_names)
-            for field in (x for x in list(default.keys()) if x not in fields_names):
-                del default[field]
+            del default['id']
         return default
