@@ -3,17 +3,23 @@
 import base64
 import gzip
 import logging
+import time
 from io import BytesIO
 from functools import wraps
+try:
+    from http import HTTPStatus
+except ImportError:
+    from http import client as HTTPStatus
 
 from werkzeug.wrappers import Request as _Request, Response
-from werkzeug.utils import cached_property
 from werkzeug.http import wsgi_to_bytes, bytes_to_wsgi
 from werkzeug.datastructures import Authorization
 from werkzeug.exceptions import abort, HTTPException
 
 from trytond import security, backend
+from trytond.exceptions import RateLimitException
 from trytond.pool import Pool
+from trytond.tools import cached_property
 from trytond.transaction import Transaction
 from trytond.config import config
 
@@ -24,6 +30,27 @@ class Request(_Request):
 
     view_args = None
 
+    def __repr__(self):
+        args = []
+        try:
+            if self.url is None or isinstance(self.url, str):
+                url = self.url
+            else:
+                url = self.url.decode(self.url_charset)
+            auth = self.authorization
+            if auth:
+                args.append("%s@%s" % (
+                        auth.get('userid', auth.username), self.remote_addr))
+            else:
+                args.append(self.remote_addr)
+            args.append("'%s'" % url)
+            args.append("[%s]" % self.method)
+            args.append("%s" % (self.rpc_method or ''))
+        except Exception:
+            args.append("(invalid WSGI environ)")
+        return "<%s %s>" % (
+            self.__class__.__name__, " ".join(filter(None, args)))
+
     @property
     def decoded_data(self):
         if self.content_encoding == 'gzip':
@@ -54,6 +81,8 @@ class Request(_Request):
 
     @cached_property
     def user_id(self):
+        if not self.view_args:
+            return None
         database_name = self.view_args.get('database_name')
         if not database_name:
             return None
@@ -66,9 +95,12 @@ class Request(_Request):
                 database_name, auth.get('userid'), auth.get('session'),
                 context=context)
         else:
-            user_id = security.login(
-                database_name, auth.username, auth, cache=False,
-                context=context)
+            try:
+                user_id = security.login(
+                    database_name, auth.username, auth, cache=False,
+                    context=context)
+            except RateLimitException:
+                abort(HTTPStatus.TOO_MANY_REQUESTS)
         return user_id
 
     @cached_property
@@ -111,6 +143,11 @@ def set_max_request_size(size):
     return decorator
 
 
+def allow_null_origin(func):
+    func.allow_null_origin = True
+    return func
+
+
 def with_pool(func):
     @wraps(func)
     def wrapper(request, database_name, *args, **kwargs):
@@ -123,27 +160,36 @@ def with_pool(func):
     return wrapper
 
 
-def with_transaction(readonly=None):
+def with_transaction(readonly=None, user=0, context=None):
     from trytond.worker import run_task
 
     def decorator(func):
         @wraps(func)
         def wrapper(request, pool, *args, **kwargs):
-            DatabaseOperationalError = backend.get('DatabaseOperationalError')
+            nonlocal user, context
             readonly_ = readonly  # can not modify non local
             if readonly_ is None:
                 if request.method in {'POST', 'PUT', 'DELETE', 'PATCH'}:
                     readonly_ = False
                 else:
                     readonly_ = True
-            context = {'_request': request.context}
-            for count in range(config.getint('database', 'retry'), -1, -1):
+            if context is None:
+                context = {}
+            else:
+                context = context.copy()
+            context['_request'] = request.context
+            if user == 'request':
+                user = request.user_id
+            retry = config.getint('database', 'retry')
+            for count in range(retry, -1, -1):
+                if count != retry:
+                    time.sleep(0.02 * (retry - count))
                 with Transaction().start(
-                        pool.database_name, 0, readonly=readonly_,
+                        pool.database_name, user, readonly=readonly_,
                         context=context) as transaction:
                     try:
                         result = func(request, pool, *args, **kwargs)
-                    except DatabaseOperationalError:
+                    except backend.DatabaseOperationalError:
                         if count and not readonly_:
                             transaction.rollback()
                             continue
@@ -176,13 +222,13 @@ def user_application(name, json=True):
                 auth_type, auth_info = authorization.split(None, 1)
                 auth_type = auth_type.lower()
             except ValueError:
-                abort(401)
+                abort(HTTPStatus.UNAUTHORIZED)
             if auth_type != b'bearer':
-                abort(403)
+                abort(HTTPStatus.FORBIDDEN)
 
             application = UserApplication.check(bytes_to_wsgi(auth_info), name)
             if not application:
-                abort(403)
+                abort(HTTPStatus.FORBIDDEN)
             transaction = Transaction()
             # TODO language
             with transaction.set_user(application.user.id), \
@@ -193,7 +239,7 @@ def user_application(name, json=True):
                     if isinstance(e, HTTPException):
                         raise
                     logger.error('%s', request, exc_info=True)
-                    abort(500, e)
+                    abort(HTTPStatus.INTERNAL_SERVER_ERROR, e)
             if not isinstance(response, Response) and json:
                 response = Response(json_.dumps(response, cls=JSONEncoder),
                     content_type='application/json')
