@@ -8,11 +8,16 @@ import logging
 from decimal import Decimal
 
 from werkzeug.wrappers import Response
-from werkzeug.utils import cached_property
-from werkzeug.exceptions import BadRequest, InternalServerError
+from werkzeug.exceptions import (
+    BadRequest, InternalServerError, Conflict, Forbidden, Locked,
+    TooManyRequests)
 
+from trytond.model.fields.dict import ImmutableDict
 from trytond.protocols.wrappers import Request
-from trytond.exceptions import TrytonException
+from trytond.exceptions import (
+    TrytonException, UserWarning, LoginException, ConcurrencyException,
+    RateLimitException, MissingDependenciesException)
+from trytond.tools import cached_property
 
 logger = logging.getLogger(__name__)
 
@@ -24,13 +29,6 @@ def dump_decimal(self, value, write):
     self.dump_struct(value, write)
 
 
-def dump_bytes(self, value, write):
-    self.write = write
-    value = client.Binary(value)
-    value.encode(self)
-    del self.write
-
-
 def dump_date(self, value, write):
     value = {'__class__': 'date',
         'year': value.year,
@@ -67,13 +65,9 @@ def dump_long(self, value, write):
 
 
 client.Marshaller.dispatch[Decimal] = dump_decimal
-client.Marshaller.dispatch[type(None)] = \
-        lambda self, value, write: write("<value><nil/></value>")
 client.Marshaller.dispatch[datetime.date] = dump_date
 client.Marshaller.dispatch[datetime.time] = dump_time
 client.Marshaller.dispatch[datetime.timedelta] = dump_timedelta
-if bytes == str:
-    client.Marshaller.dispatch[bytearray] = dump_bytes
 client.Marshaller.dispatch[int] = dump_long
 
 
@@ -87,7 +81,9 @@ def dump_struct(self, value, write, escape=client.escape):
         converted_value[k] = v
     return self.dump_struct(converted_value, write, escape=escape)
 
+
 client.Marshaller.dispatch[dict] = dump_struct
+client.Marshaller.dispatch[ImmutableDict] = dump_struct
 
 
 class XMLRPCDecoder(object):
@@ -104,6 +100,7 @@ class XMLRPCDecoder(object):
             return self.decoders[dct['__class__']](dct)
         return dct
 
+
 XMLRPCDecoder.register('date',
     lambda dct: datetime.date(dct['year'], dct['month'], dct['day']))
 XMLRPCDecoder.register('time',
@@ -125,26 +122,10 @@ def end_struct(self, data):
     self._stack[mark:] = [dct]
     self._value = 0
 
+
 client.Unmarshaller.dispatch['struct'] = end_struct
 
 
-def _end_dateTime(self, data):
-    value = client.DateTime()
-    value.decode(data)
-    value = client._datetime_type(data)
-    self.append(value)
-client.Unmarshaller.dispatch["dateTime.iso8601"] = _end_dateTime
-
-
-def _end_base64(self, data):
-    value = client.Binary()
-    value.decode(data.encode('ascii'))
-    cast = bytearray if bytes == str else bytes
-    self.append(cast(value.data))
-    self._value = 0
-client.Unmarshaller.dispatch['base64'] = _end_base64
-
-
 class XMLRequest(Request):
     parsed_content_type = 'xml'
 
@@ -153,7 +134,7 @@ class XMLRequest(Request):
         if self.parsed_content_type in self.environ.get('CONTENT_TYPE', ''):
             try:
                 # TODO replace by own loads
-                return client.loads(self.decoded_data)
+                return client.loads(self.decoded_data, use_builtin_types=True)
             except Exception:
                 raise BadRequest('Unable to read XMl request')
         else:
@@ -188,6 +169,18 @@ class XMLProtocol:
                     data, methodresponse=True, allow_none=True),
                 content_type='text/xml')
         else:
-            if isinstance(data, Exception):
+            if isinstance(data, UserWarning):
+                return Conflict(data)
+            elif isinstance(data, LoginException):
+                return Forbidden(data)
+            elif isinstance(data, ConcurrencyException):
+                return Locked(data)
+            elif isinstance(data, RateLimitException):
+                return TooManyRequests(data)
+            elif isinstance(data, MissingDependenciesException):
+                return InternalServerError(data)
+            elif isinstance(data, TrytonException):
+                return BadRequest(data)
+            elif isinstance(data, Exception):
                 return InternalServerError(data)
             return Response(data)
