@@ -4,23 +4,7 @@ import json
 import datetime
 from decimal import Decimal
 from dateutil.relativedelta import relativedelta
-from functools import reduce, wraps
-
-
-def reduced_type(types):
-    types = types.copy()
-    for k, r in [(long, int), (str, basestring), (unicode, basestring)]:
-        if k in types:
-            types.remove(k)
-            types.add(r)
-    return types
-
-
-def reduce_type(func):
-    @wraps(func)
-    def wrapper(*args, **kwargs):
-        return reduced_type(func(*args, **kwargs))
-    return wrapper
+from functools import reduce
 
 
 class PYSON(object):
@@ -36,20 +20,20 @@ class PYSON(object):
         raise NotImplementedError
 
     def __invert__(self):
-        if self.types() != set([bool]):
+        if self.types() != {bool}:
             return Not(Bool(self))
         else:
             return Not(self)
 
     def __and__(self, other):
         if (isinstance(other, PYSON)
-                and other.types() != set([bool])):
+                and other.types() != {bool}):
             other = Bool(other)
         if (isinstance(self, And)
                 and not isinstance(self, Or)):
             self._statements.append(other)
             return self
-        if self.types() != set([bool]):
+        if self.types() != {bool}:
             return And(Bool(self), other)
         else:
             return And(self, other)
@@ -58,12 +42,12 @@ class PYSON(object):
 
     def __or__(self, other):
         if (isinstance(other, PYSON)
-                and other.types() != set([bool])):
+                and other.types() != {bool}):
             other = Bool(other)
         if isinstance(self, Or):
             self._statements.append(other)
             return self
-        if self.types() != set([bool]):
+        if self.types() != {bool}:
             return Or(Bool(self), other)
         else:
             return Or(self, other)
@@ -161,15 +145,20 @@ class Eval(PYSON):
             'd': self._default,
             }
 
-    @reduce_type
     def types(self):
         if isinstance(self._default, PYSON):
             return self._default.types()
         else:
-            return set([type(self._default)])
+            return {type(self._default)}
 
     @staticmethod
     def eval(dct, context):
+        if '.' in dct['v']:
+            base, name = dct['v'].split('.', 1)
+            return Eval.eval({
+                    'v': name,
+                    'd': dct['d'],
+                    }, context.get(base) or {})
         return context.get(dct['v'], dct['d'])
 
 
@@ -178,9 +167,10 @@ class Not(PYSON):
     def __init__(self, v):
         super(Not, self).__init__()
         if isinstance(v, PYSON):
-            assert v.types() == set([bool]), 'value must be boolean'
-        else:
-            assert isinstance(v, bool), 'value must be boolean'
+            if v.types() != {bool}:
+                v = Bool(v)
+        elif not isinstance(v, bool):
+            v = bool(v)
         self._value = v
 
     @property
@@ -194,7 +184,7 @@ class Not(PYSON):
             }
 
     def types(self):
-        return set([bool])
+        return {bool}
 
     @staticmethod
     def eval(dct, context):
@@ -218,7 +208,7 @@ class Bool(PYSON):
             }
 
     def types(self):
-        return set([bool])
+        return {bool}
 
     @staticmethod
     def eval(dct, context):
@@ -230,13 +220,12 @@ class And(PYSON):
     def __init__(self, *statements, **kwargs):
         super(And, self).__init__()
         statements = list(statements) + kwargs.get('s', [])
-        for statement in statements:
+        for i, statement in enumerate(list(statements)):
             if isinstance(statement, PYSON):
-                assert statement.types() == set([bool]), \
-                    'statement must be boolean'
-            else:
-                assert isinstance(statement, bool), \
-                    'statement must be boolean'
+                if statement.types() != {bool}:
+                    statements[i] = Bool(statement)
+            elif not isinstance(statement, bool):
+                statements[i] = bool(statement)
         assert len(statements) >= 2, 'must have at least 2 statements'
         self._statements = statements
 
@@ -251,7 +240,7 @@ class And(PYSON):
             }
 
     def types(self):
-        return set([bool])
+        return {bool}
 
     @staticmethod
     def eval(dct, context):
@@ -278,11 +267,11 @@ class Equal(PYSON):
         if isinstance(statement1, PYSON):
             types1 = statement1.types()
         else:
-            types1 = reduced_type(set([type(s1)]))
+            types1 = {type(s1)}
         if isinstance(statement2, PYSON):
             types2 = statement2.types()
         else:
-            types2 = reduced_type(set([type(s2)]))
+            types2 = {type(s2)}
         assert types1 == types2, 'statements must have the same type'
         self._statement1 = statement1
         self._statement2 = statement2
@@ -299,7 +288,7 @@ class Equal(PYSON):
             }
 
     def types(self):
-        return set([bool])
+        return {bool}
 
     @staticmethod
     def eval(dct, context):
@@ -313,15 +302,20 @@ class Greater(PYSON):
         super(Greater, self).__init__()
         for i in (statement1, statement2):
             if isinstance(i, PYSON):
-                assert i.types().issubset({int, long, float, type(None)}), \
-                    'statement must be an integer or a float'
+                assert i.types().issubset({
+                        int, float, type(None),
+                        datetime.datetime, datetime.date}), \
+                    'statement must be an integer, float, date or datetime'
             else:
-                assert isinstance(i, (int, long, float, type(None))), \
-                    'statement must be an integer or a float'
+                assert isinstance(i, (
+                        int, float, type(None),
+                        datetime.datetime, datetime.date)), \
+                    'statement must be an integer, float, date or datetime'
         if isinstance(equal, PYSON):
-            assert equal.types() == set([bool])
-        else:
-            assert isinstance(equal, bool)
+            if equal.types() != {bool}:
+                equal = Bool(equal)
+        elif not isinstance(equal, bool):
+            equal = bool(equal)
         self._statement1 = statement1
         self._statement2 = statement2
         self._equal = equal
@@ -339,20 +333,28 @@ class Greater(PYSON):
             }
 
     def types(self):
-        return set([bool])
+        return {bool}
 
     @staticmethod
     def _convert(dct):
         for i in ('s1', 's2'):
             if dct[i] is None:
                 dct[i] = 0.0
-            if not isinstance(dct[i], (int, long, float)):
+            if not isinstance(dct[i], (int, float)):
                 dct = dct.copy()
-                dct[i] = float(dct[i])
+                stmt = dct[i]
+                if isinstance(stmt, datetime.datetime):
+                    stmt = stmt.timestamp()
+                elif isinstance(stmt, datetime.date):
+                    time = datetime.time(0, 0)
+                    stmt = datetime.datetime.combine(stmt, time).timestamp()
+                dct[i] = float(stmt)
         return dct
 
     @staticmethod
     def eval(dct, context):
+        if dct['s1'] is None or dct['s2'] is None:
+            return False
         dct = Greater._convert(dct)
         if dct['e']:
             return dct['s1'] >= dct['s2']
@@ -369,6 +371,8 @@ class Less(Greater):
 
     @staticmethod
     def eval(dct, context):
+        if dct['s1'] is None or dct['s2'] is None:
+            return False
         dct = Less._convert(dct)
         if dct['e']:
             return dct['s1'] <= dct['s2']
@@ -382,18 +386,18 @@ class If(PYSON):
         condition, then_statement, else_statement = c, t, e
         super(If, self).__init__()
         if isinstance(condition, PYSON):
-            assert condition.types() == set([bool]), \
-                'condition must be boolean'
-        else:
-            assert isinstance(condition, bool), 'condition must be boolean'
+            if condition.types() != {bool}:
+                condition = Bool(condition)
+        elif not isinstance(condition, bool):
+            condition = bool(condition)
         if isinstance(then_statement, PYSON):
             then_types = then_statement.types()
         else:
-            then_types = reduced_type(set([type(then_statement)]))
+            then_types = {type(then_statement)}
         if isinstance(else_statement, PYSON):
             else_types = else_statement.types()
         else:
-            else_types = reduced_type(set([type(else_statement)]))
+            else_types = {type(else_statement)}
         assert then_types == else_types, \
             'then and else statements must be the same type'
         self._condition = condition
@@ -412,12 +416,11 @@ class If(PYSON):
             'e': self._else_statement,
             }
 
-    @reduce_type
     def types(self):
         if isinstance(self._then_statement, PYSON):
             return self._then_statement.types()
         else:
-            return set([type(self._then_statement)])
+            return {type(self._then_statement)}
 
     @staticmethod
     def eval(dct, context):
@@ -433,14 +436,14 @@ class Get(PYSON):
         obj, key, default = v, k, d
         super(Get, self).__init__()
         if isinstance(obj, PYSON):
-            assert obj.types() == set([dict]), 'obj must be a dict'
+            assert obj.types() == {dict}, 'obj must be a dict'
         else:
             assert isinstance(obj, dict), 'obj must be a dict'
         self._obj = obj
         if isinstance(key, PYSON):
-            assert key.types() == set([basestring]), 'key must be a string'
+            assert key.types() == {str}, 'key must be a string'
         else:
-            assert isinstance(key, basestring), 'key must be a string'
+            assert isinstance(key, str), 'key must be a string'
         self._key = key
         self._default = default
 
@@ -456,12 +459,11 @@ class Get(PYSON):
             'd': self._default,
             }
 
-    @reduce_type
     def types(self):
         if isinstance(self._default, PYSON):
             return self._default.types()
         else:
-            return set([type(self._default)])
+            return {type(self._default)}
 
     @staticmethod
     def eval(dct, context):
@@ -474,20 +476,20 @@ class In(PYSON):
         key, obj = k, v
         super(In, self).__init__()
         if isinstance(key, PYSON):
-            assert key.types().issubset(set([basestring, int])), \
+            assert key.types().issubset({str, int}), \
                 'key must be a string or an integer or a long'
         else:
-            assert isinstance(key, (basestring, int, long)), \
+            assert isinstance(key, (str, int)), \
                 'key must be a string or an integer or a long'
         if isinstance(obj, PYSON):
-            assert obj.types().issubset(set([dict, list])), \
+            assert obj.types().issubset({dict, list}), \
                 'obj must be a dict or a list'
-            if obj.types() == set([dict]):
-                assert isinstance(key, basestring), 'key must be a string'
+            if obj.types() == {dict}:
+                assert isinstance(key, str), 'key must be a string'
         else:
             assert isinstance(obj, (dict, list))
             if isinstance(obj, dict):
-                assert isinstance(key, basestring), 'key must be a string'
+                assert isinstance(key, str), 'key must be a string'
         self._key = key
         self._obj = obj
 
@@ -503,7 +505,7 @@ class In(PYSON):
             }
 
     def types(self):
-        return set([bool])
+        return {bool}
 
     @staticmethod
     def eval(dct, context):
@@ -513,7 +515,7 @@ class In(PYSON):
 class Date(PYSON):
 
     def __init__(self, year=None, month=None, day=None,
-            delta_years=0, delta_months=0, delta_days=0, **kwargs):
+            delta_years=0, delta_months=0, delta_days=0, start=None, **kwargs):
         year = kwargs.get('y', year)
         month = kwargs.get('M', month)
         day = kwargs.get('d', day)
@@ -523,10 +525,10 @@ class Date(PYSON):
         super(Date, self).__init__()
         for i in (year, month, day, delta_years, delta_months, delta_days):
             if isinstance(i, PYSON):
-                assert i.types().issubset(set([int, long, type(None)])), \
+                assert i.types().issubset({int, type(None)}), \
                     '%s must be an integer or None' % (i,)
             else:
-                assert isinstance(i, (int, long, type(None))), \
+                assert isinstance(i, (int, type(None))), \
                     '%s must be an integer or None' % (i,)
         self._year = year
         self._month = month
@@ -534,11 +536,13 @@ class Date(PYSON):
         self._delta_years = delta_years
         self._delta_months = delta_months
         self._delta_days = delta_days
+        self._start = start
 
     @property
     def __repr_params__(self):
         return (self._year, self._month, self._day,
-            self._delta_years, self._delta_months, self._delta_days)
+            self._delta_years, self._delta_months, self._delta_days,
+            self._start)
 
     def pyson(self):
         return {
@@ -549,14 +553,20 @@ class Date(PYSON):
             'dy': self._delta_years,
             'dM': self._delta_months,
             'dd': self._delta_days,
+            'start': self._start,
             }
 
     def types(self):
-        return set([datetime.date])
+        return {datetime.date}
 
     @staticmethod
     def eval(dct, context):
-        return datetime.date.today() + relativedelta(
+        today = dct.get('start')
+        if isinstance(today, datetime.datetime):
+            today = today.date()
+        if not isinstance(today, datetime.date):
+            today = datetime.date.today()
+        return today + relativedelta(
             year=dct['y'],
             month=dct['M'],
             day=dct['d'],
@@ -572,7 +582,7 @@ class DateTime(Date):
             hour=None, minute=None, second=None, microsecond=None,
             delta_years=0, delta_months=0, delta_days=0,
             delta_hours=0, delta_minutes=0, delta_seconds=0,
-            delta_microseconds=0, **kwargs):
+            delta_microseconds=0, start=None, **kwargs):
         hour = kwargs.get('h', hour)
         minute = kwargs.get('m', minute)
         second = kwargs.get('s', second)
@@ -583,14 +593,14 @@ class DateTime(Date):
         delta_microseconds = kwargs.get('dms', delta_microseconds)
         super(DateTime, self).__init__(year=year, month=month, day=day,
                 delta_years=delta_years, delta_months=delta_months,
-                delta_days=delta_days, **kwargs)
+                delta_days=delta_days, start=start, **kwargs)
         for i in (hour, minute, second, microsecond,
                 delta_hours, delta_minutes, delta_seconds, delta_microseconds):
             if isinstance(i, PYSON):
-                assert i.types() == set([int, type(None)]), \
+                assert i.types() == {int, type(None)}, \
                     '%s must be an integer or None' % (i,)
             else:
-                assert isinstance(i, (int, long, type(None))), \
+                assert isinstance(i, (int, type(None))), \
                     '%s must be an integer or None' % (i,)
         self._hour = hour
         self._minute = minute
@@ -606,9 +616,10 @@ class DateTime(Date):
         date_params = super(DateTime, self).__repr_params__
         return (date_params[:3]
             + (self._hour, self._minute, self._second, self._microsecond)
-            + date_params[3:]
+            + date_params[3:-1]
             + (self._delta_hours, self._delta_minutes, self._delta_seconds,
-                self._delta_microseconds))
+                self._delta_microseconds)
+            + date_params[-1:])
 
     def pyson(self):
         res = super(DateTime, self).pyson()
@@ -624,11 +635,17 @@ class DateTime(Date):
         return res
 
     def types(self):
-        return set([datetime.datetime])
+        return {datetime.datetime}
 
     @staticmethod
     def eval(dct, context):
-        return datetime.datetime.now() + relativedelta(
+        now = dct.get('start')
+        if (isinstance(now, datetime.date)
+                and not isinstance(now, datetime.datetime)):
+            now = datetime.datetime.combine(now, datetime.time())
+        if not isinstance(now, datetime.datetime):
+            now = datetime.datetime.utcnow()
+        return now + relativedelta(
             year=dct['y'],
             month=dct['M'],
             day=dct['d'],
@@ -646,15 +663,53 @@ class DateTime(Date):
             )
 
 
+class TimeDelta(PYSON):
+
+    def __init__(self, days=0, seconds=0, microseconds=0):
+        for i in [days, seconds, microseconds]:
+            if isinstance(i, PYSON):
+                assert i.types().issubset({int, float}), \
+                        '%s must be an integer' % (i,)
+            else:
+                assert isinstance(i, (int, float)), \
+                        '%s must be an integer' % (i,)
+        self._days = days
+        self._seconds = seconds
+        self._microseconds = microseconds
+
+    @property
+    def __repr_params__(self):
+        return self._days, self._seconds, self._microseconds
+
+    def pyson(self):
+        return {
+            '__class__': 'TimeDelta',
+            'd': self._days,
+            's': self._seconds,
+            'm': self._microseconds,
+            }
+
+    def types(self):
+        return {datetime.timedelta}
+
+    @staticmethod
+    def eval(dct, context):
+        return datetime.timedelta(
+            days=dct['d'],
+            seconds=dct['s'],
+            microseconds=dct['m'],
+            )
+
+
 class Len(PYSON):
 
     def __init__(self, v):
         super(Len, self).__init__()
         if isinstance(v, PYSON):
-            assert v.types().issubset(set([dict, list, basestring])), \
+            assert v.types().issubset({dict, list, str}), \
                 'value must be a dict or a list or a string'
         else:
-            assert isinstance(v, (dict, list, basestring)), \
+            assert isinstance(v, (dict, list, str)), \
                 'value must be a dict or list or a string'
         self._value = v
 
@@ -669,7 +724,7 @@ class Len(PYSON):
             }
 
     def types(self):
-        return set([int])
+        return {int}
 
     @staticmethod
     def eval(dct, context):
@@ -694,7 +749,8 @@ class Id(PYSON):
         return ModelData.get_id(self._module, self._fs_id)
 
     def types(self):
-        return set([int])
+        return {int}
+
 
 CONTEXT = {
     'Eval': Eval,
@@ -710,6 +766,9 @@ CONTEXT = {
     'In': In,
     'Date': Date,
     'DateTime': DateTime,
+    'TimeDelta': TimeDelta,
     'Len': Len,
     'Id': Id,
+    'true': True,
+    'false': False,
 }
