@@ -4,10 +4,11 @@ import logging
 import smtplib
 from email.message import Message
 from email.utils import formatdate
+from email.mime.text import MIMEText
 from urllib.parse import parse_qs, unquote_plus
 
-from .config import config, parse_uri
-from .transaction import Transaction
+from trytond.config import config, parse_uri
+from trytond.transaction import Transaction
 
 __all__ = ['sendmail_transactional', 'sendmail', 'SMTPDataManager']
 logger = logging.getLogger(__name__)
@@ -24,9 +25,11 @@ def sendmail_transactional(
     datamanager.put(from_addr, to_addrs, msg)
 
 
-def sendmail(from_addr, to_addrs, msg, server=None):
+def sendmail(from_addr, to_addrs, msg, server=None, strict=False):
     if server is None:
-        server = get_smtp_server()
+        server = get_smtp_server(strict=strict)
+        if not server:
+            return
         quit = True
     else:
         quit = False
@@ -34,7 +37,9 @@ def sendmail(from_addr, to_addrs, msg, server=None):
         msg['Date'] = formatdate()
     try:
         senderrs = server.sendmail(from_addr, to_addrs, msg.as_string())
-    except smtplib.SMTPException:
+    except Exception:
+        if strict:
+            raise
         logger.error('fail to send email', exc_info=True)
     else:
         if senderrs:
@@ -43,7 +48,17 @@ def sendmail(from_addr, to_addrs, msg, server=None):
         server.quit()
 
 
-def get_smtp_server(uri=None):
+def send_test_email(to_addrs, server=None):
+    from_ = config.get('email', 'from')
+    msg = MIMEText('Success!\nYour email settings work correctly.')
+    msg['From'] = from_
+    msg['To'] = to_addrs
+    msg['Subject'] = 'Tryton test email'
+    sendmail(
+        config.get('email', 'from'), to_addrs, msg, server=server, strict=True)
+
+
+def get_smtp_server(uri=None, strict=False):
     if uri is None:
         uri = config.get('email', 'uri')
     uri = parse_uri(uri)
@@ -53,9 +68,16 @@ def get_smtp_server(uri=None):
         for key, value in parse_qs(uri.query, strict_parsing=True).items():
             extra[key] = cast.get(key, lambda a: a)(value[0])
     if uri.scheme.startswith('smtps'):
-        server = smtplib.SMTP_SSL(uri.hostname, uri.port, **extra)
+        connector = smtplib.SMTP_SSL
     else:
-        server = smtplib.SMTP(uri.hostname, uri.port, **extra)
+        connector = smtplib.SMTP
+    try:
+        server = connector(uri.hostname, uri.port, **extra)
+    except Exception:
+        if strict:
+            raise
+        logger.error('fail to connect to %s', uri, exc_info=True)
+        return
 
     if 'tls' in uri.scheme:
         server.starttls()
@@ -69,8 +91,9 @@ def get_smtp_server(uri=None):
 
 class SMTPDataManager(object):
 
-    def __init__(self, uri=None):
+    def __init__(self, uri=None, strict=False):
         self.uri = uri
+        self.strict = strict
         self.queue = []
         self._server = None
 
@@ -81,7 +104,7 @@ class SMTPDataManager(object):
     def __eq__(self, other):
         if not isinstance(other, SMTPDataManager):
             return NotImplemented
-        return self.uri == other.uri
+        return (self.uri == other.uri) and (self.strict == other.strict)
 
     def abort(self, trans):
         self._finish()
@@ -94,7 +117,7 @@ class SMTPDataManager(object):
 
     def tpc_vote(self, trans):
         if self._server is None:
-            self._server = get_smtp_server(self.uri)
+            self._server = get_smtp_server(self.uri, strict=self.strict)
 
     def tpc_finish(self, trans):
         if self._server is not None:
