@@ -93,6 +93,20 @@ class CopyMany2ManyReferenceRelation(ModelSQL):
         'Many2ManyReference Target')
 
 
+class CopyBinary(ModelSQL):
+    "Copy Binary"
+    __name__ = 'test.copy.binary'
+    binary = fields.Binary("Binary")
+    binary_id = fields.Binary("Binary with ID", file_id='file_id')
+    file_id = fields.Char("Binary ID")
+
+
+class CopyTranslate(ModelSQL):
+    "Copy Translate"
+    __name__ = 'test.copy.translate'
+    name = fields.Char("Name", translate=True)
+
+
 def register(module):
     Pool.register(
         Copy,
@@ -106,4 +120,6 @@ def register(module):
         CopyMany2ManyReference,
         CopyMany2ManyReferenceTarget,
         CopyMany2ManyReferenceRelation,
+        CopyBinary,
+        CopyTranslate,
         module=module, type_='model')
