@@ -2,11 +2,7 @@
 # this repository contains the full copyright notices and license terms.
 "Test for export_data"
 from trytond.model import ModelSQL, fields
-from trytond.pool import PoolMeta
-
-__all__ = [
-    'ExportDataTarget', 'ExportData', 'ExportDataTarget2',
-    'ExportDataRelation']
+from trytond.pool import PoolMeta, Pool
 
 
 class ExportDataTarget(ModelSQL):
@@ -26,6 +22,7 @@ class ExportData(ModelSQL):
     text = fields.Text('Text')
     date = fields.Date('Date')
     datetime = fields.DateTime('DateTime')
+    timedelta = fields.TimeDelta('TimeDelta')
     selection = fields.Selection([
             (None, ''),
             ('select1', 'Select 1'),
@@ -43,9 +40,8 @@ class ExportData(ModelSQL):
             ])
 
 
-class ExportDataTarget2:
+class ExportDataTarget2(metaclass=PoolMeta):
     'Export Date Target'
-    __metaclass__ = PoolMeta
     __name__ = 'test.export_data.target'
     one2many = fields.Many2One('test.export_data', 'Export Data')
 
@@ -55,3 +51,12 @@ class ExportDataRelation(ModelSQL):
     __name__ = 'test.export_data.relation'
     many2many = fields.Many2One('test.export_data', 'Export Data')
     target = fields.Many2One('test.export_data.target', 'Target')
+
+
+def register(module):
+    Pool.register(
+        ExportDataTarget,
+        ExportData,
+        ExportDataTarget2,
+        ExportDataRelation,
+        module=module, type_='model')
