@@ -1,14 +1,6 @@
 # -*- coding: utf-8 -*-
 # This file is part of Tryton.  The COPYRIGHT file at the top level of
 # this repository contains the full copyright notices and license terms.
-import sys
-try:
-    import cdecimal
-    if 'decimal' not in sys.modules:
-        sys.modules['decimal'] = cdecimal
-except ImportError:
-    import decimal
-    sys.modules['cdecimal'] = decimal
 import unittest
 from decimal import Decimal
 import datetime
@@ -168,7 +160,7 @@ class ExportDataTestCase(unittest.TestCase):
         export1, = ExportData.create([{
                     'date': datetime.date(2010, 1, 1),
                     }])
-        self.assert_(ExportData.export_data([export1],
+        self.assertTrue(ExportData.export_data([export1],
             ['date']) == [[datetime.date(2010, 1, 1)]])
 
         export2, = ExportData.create([{
@@ -205,6 +197,33 @@ class ExportDataTestCase(unittest.TestCase):
             ExportData.export_data([export1, export2], ['datetime']),
             [[datetime.datetime(2010, 1, 1, 12, 0, 0)], ['']])
 
+    @with_transaction()
+    def test_timedelta(self):
+        'Test export_data timedelta'
+        pool = Pool()
+        ExportData = pool.get('test.export_data')
+
+        export1, = ExportData.create([{
+                    'timedelta': datetime.timedelta(
+                        hours=1, minutes=20, seconds=30.45),
+                    }])
+        self.assertEqual(
+            ExportData.export_data([export1], ['timedelta']),
+            [[datetime.timedelta(hours=1, minutes=20, seconds=30.45)]])
+
+        export2, = ExportData.create([{
+                    'timedelta': None,
+                    }])
+        self.assertEqual(
+            ExportData.export_data([export2], ['timedelta']),
+            [['']])
+
+        self.assertEqual(
+            ExportData.export_data([export1, export2], ['timedelta']), [
+                [datetime.timedelta(hours=1, minutes=20, seconds=30.45)],
+                [''],
+                ])
+
     @with_transaction()
     def test_selection(self):
         'Test export_data selection'
@@ -368,6 +387,23 @@ class ExportDataTestCase(unittest.TestCase):
             ExportData.export_data([export1], ['reference.translated']),
             [["Target"]])
 
+    @with_transaction()
+    def test_domain(self):
+        "Test export data with domain"
+        pool = Pool()
+        ExportData = pool.get('test.export_data')
+
+        ExportData.create([{
+                    'boolean': True,
+                    }, {
+                    'boolean': False,
+                    }])
+
+        self.assertEqual(
+            ExportData.export_data_domain(
+                [('boolean', '=', True)], ['boolean']),
+            [[True]])
+
 
 def suite():
     return unittest.TestLoader().loadTestsFromTestCase(ExportDataTestCase)
