@@ -4,6 +4,8 @@
 
 import unittest
 import datetime
+import sys
+
 from decimal import Decimal
 from trytond import pyson
 
@@ -109,8 +111,9 @@ class PYSONTestCase(unittest.TestCase):
             's': [True, False],
             })
 
-        self.assertRaises(AssertionError, pyson.And, True)
-        self.assertRaises(AssertionError, pyson.And)
+        if not sys.flags.optimize:
+            self.assertRaises(AssertionError, pyson.And, True)
+            self.assertRaises(AssertionError, pyson.And)
 
         self.assertEqual(pyson.And(True, False).types(), set([bool]))
 
@@ -151,8 +154,9 @@ class PYSONTestCase(unittest.TestCase):
             's': [True, False],
             })
 
-        self.assertRaises(AssertionError, pyson.Or, True)
-        self.assertRaises(AssertionError, pyson.Or)
+        if not sys.flags.optimize:
+            self.assertRaises(AssertionError, pyson.Or, True)
+            self.assertRaises(AssertionError, pyson.Or)
 
         self.assertEqual(pyson.Or(True, False).types(), set([bool]))
 
@@ -194,7 +198,8 @@ class PYSONTestCase(unittest.TestCase):
             's2': 'test',
             })
 
-        self.assertRaises(AssertionError, pyson.Equal, 'test', True)
+        if not sys.flags.optimize:
+            self.assertRaises(AssertionError, pyson.Equal, 'test', True)
 
         self.assertEqual(pyson.Equal('test', 'test').types(), set([bool]))
 
@@ -216,8 +221,15 @@ class PYSONTestCase(unittest.TestCase):
             'e': False,
             })
 
-        self.assertRaises(AssertionError, pyson.Greater, 'test', 0)
-        self.assertRaises(AssertionError, pyson.Greater, 1, 'test')
+        if not sys.flags.optimize:
+            self.assertRaises(AssertionError, pyson.Greater, 'test', 0)
+            self.assertRaises(AssertionError, pyson.Greater, 1, 'test')
+            self.assertRaises(
+                AssertionError, pyson.Greater, pyson.Eval('foo'), 0)
+            self.assertRaises(AssertionError, pyson.Greater,
+                'test', pyson.DateTime())
+            self.assertRaises(AssertionError, pyson.Greater,
+                pyson.DateTime(), 'test')
 
         self.assertEqual(pyson.Greater(1, 0).types(), set([bool]))
 
@@ -243,10 +255,58 @@ class PYSONTestCase(unittest.TestCase):
         self.assertFalse(pyson.PYSONDecoder().decode(eval))
 
         eval = pyson.PYSONEncoder().encode(pyson.Greater(1, None))
-        self.assertTrue(pyson.PYSONDecoder().decode(eval))
+        self.assertFalse(pyson.PYSONDecoder().decode(eval))
 
         self.assertEqual(repr(pyson.Greater(1, 0)), 'Greater(1, 0, False)')
 
+        eval = pyson.PYSONEncoder().encode(pyson.Greater(
+                pyson.DateTime(2020, 1, 1, 0, 0, 0, 0),
+                datetime.date(2020, 1, 2)))
+        self.assertFalse(pyson.PYSONDecoder().decode(eval))
+
+        eval = pyson.PYSONEncoder().encode(pyson.Greater(
+                pyson.DateTime(2020, 1, 1, 0, 0, 0, 1),
+                datetime.date(2020, 1, 1)))
+        self.assertTrue(pyson.PYSONDecoder().decode(eval))
+
+        eval = pyson.PYSONEncoder().encode(pyson.Greater(
+                pyson.DateTime(2020, 1, 1, 0, 0, 0, 0),
+                datetime.date(2020, 1, 1), True))
+        self.assertTrue(pyson.PYSONDecoder().decode(eval))
+
+        eval = pyson.PYSONEncoder().encode(pyson.Greater(
+                pyson.DateTime(2020, 1, 1, 0, 0, 0, 0),
+                pyson.DateTime(2020, 1, 1, 0, 0, 0, 0)))
+        self.assertFalse(pyson.PYSONDecoder().decode(eval))
+
+        eval = pyson.PYSONEncoder().encode(pyson.Greater(
+                pyson.Date(2020, 1, 1),
+                datetime.date(2020, 1, 1)))
+        self.assertFalse(pyson.PYSONDecoder().decode(eval))
+
+        eval = pyson.PYSONEncoder().encode(pyson.Greater(
+                pyson.Date(2020, 1, 1),
+                pyson.DateTime(2020, 1, 1, 0, 0, 0, 1)))
+        self.assertFalse(pyson.PYSONDecoder().decode(eval))
+
+        eval = pyson.PYSONEncoder().encode(pyson.Greater(
+                pyson.DateTime(2020, 1, 1, 0, 0, 1),
+                pyson.Date(2020, 1, 1), True))
+        self.assertTrue(pyson.PYSONDecoder().decode(eval))
+
+        eval = pyson.PYSONEncoder().encode(pyson.Greater(
+                pyson.DateTime(2020, 1, 1, 0, 0, 0, 0),
+                pyson.Date(2020, 1, 1), True))
+        self.assertTrue(pyson.PYSONDecoder().decode(eval))
+
+        eval = pyson.PYSONEncoder().encode(pyson.Greater(
+                pyson.DateTime(2020, 1, 1, 0, 0, 0, 0), 90000))
+        self.assertTrue(pyson.PYSONDecoder().decode(eval))
+
+        eval = pyson.PYSONEncoder().encode(
+            pyson.Greater(pyson.Eval('i', 0), 0))
+        self.assertTrue(pyson.PYSONDecoder({'i': 1}).decode(eval))
+
     def test_Less(self):
         'Test pyson.Less'
         self.assertEqual(pyson.Less(0, 1).pyson(), {
@@ -256,8 +316,13 @@ class PYSONTestCase(unittest.TestCase):
             'e': False,
             })
 
-        self.assertRaises(AssertionError, pyson.Less, 'test', 1)
-        self.assertRaises(AssertionError, pyson.Less, 0, 'test')
+        if not sys.flags.optimize:
+            self.assertRaises(AssertionError, pyson.Less, 'test', 1)
+            self.assertRaises(AssertionError, pyson.Less, 0, 'test')
+            self.assertRaises(AssertionError, pyson.Less,
+                'test', pyson.DateTime())
+            self.assertRaises(AssertionError, pyson.Less,
+                pyson.DateTime(), 'test')
 
         self.assertEqual(pyson.Less(0, 1).types(), set([bool]))
 
@@ -280,13 +345,47 @@ class PYSONTestCase(unittest.TestCase):
         self.assertTrue(pyson.PYSONDecoder().decode(eval))
 
         eval = pyson.PYSONEncoder().encode(pyson.Less(None, 1))
-        self.assertTrue(pyson.PYSONDecoder().decode(eval))
+        self.assertFalse(pyson.PYSONDecoder().decode(eval))
 
         eval = pyson.PYSONEncoder().encode(pyson.Less(1, None))
         self.assertFalse(pyson.PYSONDecoder().decode(eval))
 
         self.assertEqual(repr(pyson.Less(0, 1)), 'Less(0, 1, False)')
 
+        eval = pyson.PYSONEncoder().encode(pyson.Less(
+                pyson.DateTime(2020, 1, 1, 0, 0, 0, 0),
+                datetime.date(2020, 1, 1)))
+        self.assertFalse(pyson.PYSONDecoder().decode(eval))
+
+        eval = pyson.PYSONEncoder().encode(pyson.Less(
+                pyson.DateTime(2020, 1, 1, 0, 0, 0, 0),
+                datetime.date(2020, 1, 2), True))
+        self.assertTrue(pyson.PYSONDecoder().decode(eval))
+
+        eval = pyson.PYSONEncoder().encode(pyson.Less(
+                pyson.DateTime(2020, 1, 1, 0, 0, 0, 0),
+                pyson.DateTime(2020, 1, 1, 0, 0, 0, 0)))
+        self.assertFalse(pyson.PYSONDecoder().decode(eval))
+
+        eval = pyson.PYSONEncoder().encode(pyson.Less(
+                pyson.Date(2020, 1, 1),
+                datetime.date(2020, 1, 2)))
+        self.assertTrue(pyson.PYSONDecoder().decode(eval))
+
+        eval = pyson.PYSONEncoder().encode(pyson.Less(
+                pyson.Date(2020, 1, 1),
+                pyson.DateTime(2020, 1, 1, 0, 0, 0, 1)))
+        self.assertTrue(pyson.PYSONDecoder().decode(eval))
+
+        eval = pyson.PYSONEncoder().encode(pyson.Less(
+                pyson.DateTime(2020, 1, 1, 0, 0, 0, 0),
+                pyson.Date(2020, 1, 1), True))
+        self.assertTrue(pyson.PYSONDecoder().decode(eval))
+
+        eval = pyson.PYSONEncoder().encode(pyson.Less(
+                pyson.DateTime(2020, 1, 1, 0, 0, 0, 0), 90000))
+        self.assertFalse(pyson.PYSONDecoder().decode(eval))
+
     def test_If(self):
         'Test pyson.If'
         self.assertEqual(pyson.If(True, 'foo', 'bar').pyson(), {
@@ -296,7 +395,8 @@ class PYSONTestCase(unittest.TestCase):
             'e': 'bar',
             })
 
-        self.assertRaises(AssertionError, pyson.If, True, 'foo', False)
+        if not sys.flags.optimize:
+            self.assertRaises(AssertionError, pyson.If, True, 'foo', False)
 
         self.assertEqual(pyson.If(True, 'foo', 'bar').types(),
             set([str]))
@@ -320,8 +420,10 @@ class PYSONTestCase(unittest.TestCase):
             'd': 'default',
             })
 
-        self.assertRaises(AssertionError, pyson.Get, 'test', 'foo', 'default')
-        self.assertRaises(AssertionError, pyson.Get, {}, 1, 'default')
+        if not sys.flags.optimize:
+            self.assertRaises(
+                AssertionError, pyson.Get, 'test', 'foo', 'default')
+            self.assertRaises(AssertionError, pyson.Get, {}, 1, 'default')
 
         self.assertEqual(pyson.Get({}, 'foo', 'default').types(),
             set([str]))
@@ -350,8 +452,9 @@ class PYSONTestCase(unittest.TestCase):
             'v': {'foo': 'bar'},
             })
 
-        self.assertRaises(AssertionError, pyson.In, object(), {})
-        self.assertRaises(AssertionError, pyson.In, 'test', 'foo')
+        if not sys.flags.optimize:
+            self.assertRaises(AssertionError, pyson.In, object(), {})
+            self.assertRaises(AssertionError, pyson.In, 'test', 'foo')
 
         self.assertEqual(pyson.In('foo', {}).types(), set([bool]))
 
@@ -397,21 +500,23 @@ class PYSONTestCase(unittest.TestCase):
             'd': 12,
             'dy': -1,
             'dM': 12,
-            'dd': -7
+            'dd': -7,
+            'start': None,
             })
 
-        self.assertRaises(AssertionError, pyson.Date, 'test', 1, 12, -1, 12,
-                -7)
-        self.assertRaises(AssertionError, pyson.Date, 2010, 'test', 12, -1, 12,
-                -7)
-        self.assertRaises(AssertionError, pyson.Date, 2010, 1, 'test', -1, 12,
-                -7)
-        self.assertRaises(AssertionError, pyson.Date, 2010, 1, 12, 'test', 12,
-                -7)
-        self.assertRaises(AssertionError, pyson.Date, 2010, 1, 12, -1, 'test',
-                -7)
-        self.assertRaises(AssertionError, pyson.Date, 2010, 1, 12, -1, 12,
-                'test')
+        if not sys.flags.optimize:
+            self.assertRaises(
+                AssertionError, pyson.Date, 'test', 1, 12, -1, 12, -7)
+            self.assertRaises(
+                AssertionError, pyson.Date, 2010, 'test', 12, -1, 12, -7)
+            self.assertRaises(
+                AssertionError, pyson.Date, 2010, 1, 'test', -1, 12, -7)
+            self.assertRaises(
+                AssertionError, pyson.Date, 2010, 1, 12, 'test', 12, -7)
+            self.assertRaises(
+                AssertionError, pyson.Date, 2010, 1, 12, -1, 'test', -7)
+            self.assertRaises(
+                AssertionError, pyson.Date, 2010, 1, 12, -1, 12, 'test')
 
         self.assertEqual(pyson.Date(2010, 1, 12, -1, 12, -7).types(),
             set([datetime.date]))
@@ -441,7 +546,33 @@ class PYSONTestCase(unittest.TestCase):
             datetime.date(2010, 2, 22))
 
         self.assertEqual(repr(pyson.Date(2010, 1, 12, -1, 12, -7)),
-            'Date(2010, 1, 12, -1, 12, -7)')
+            'Date(2010, 1, 12, -1, 12, -7, None)')
+
+    def test_Date_start(self):
+        "Test Date with start"
+        eval = pyson.PYSONEncoder().encode(pyson.Date(
+                start=pyson.Eval('date')))
+
+        date = datetime.date(2000, 1, 1)
+        self.assertEqual(pyson.PYSONDecoder(
+                {'date': date}).decode(eval), date)
+
+    def test_Date_start_datetime(self):
+        "Test Date with start as datetime"
+        eval = pyson.PYSONEncoder().encode(pyson.Date(
+                start=pyson.Eval('datetime')))
+
+        datetime_ = datetime.datetime(2000, 1, 1, 12, 00)
+        self.assertEqual(pyson.PYSONDecoder(
+                {'datetime': datetime_}).decode(eval), datetime_.date())
+
+    def test_Date_start_invalid(self):
+        "Test Date with invalid start"
+        eval = pyson.PYSONEncoder().encode(pyson.Date(
+                start=pyson.Eval('foo')))
+
+        self.assertEqual(pyson.PYSONDecoder(
+                {'foo': 'bar'}).decode(eval), datetime.date.today())
 
     def test_DateTime(self):
         'Test pyson.DateTime'
@@ -462,36 +593,52 @@ class PYSONTestCase(unittest.TestCase):
                 'dm': 15,
                 'ds': 30,
                 'dms': 1,
+                'start': None,
                 })
 
-        self.assertRaises(AssertionError, pyson.DateTime, 'test', 1, 12, 10,
-                30, 20, 0, -1, 12, -7, 2, 15, 30, 1)
-        self.assertRaises(AssertionError, pyson.DateTime, 2010, 'test', 12, 10,
-                30, 20, 0, -1, 12, -7, 2, 15, 30, 1)
-        self.assertRaises(AssertionError, pyson.DateTime, 2010, 1, 'test', 10,
-                30, 20, 0, -1, 12, -7, 2, 15, 30, 1)
-        self.assertRaises(AssertionError, pyson.DateTime, 2010, 1, 12, 'test',
-                30, 20, 0, -1, 12, -7, 2, 15, 30, 1)
-        self.assertRaises(AssertionError, pyson.DateTime, 2010, 1, 12, 10,
-                'test', 20, 0, -1, 12, -7, 2, 15, 30, 1)
-        self.assertRaises(AssertionError, pyson.DateTime, 2010, 1, 12, 10, 30,
-                'test', 0, -1, 12, -7, 2, 15, 30, 1)
-        self.assertRaises(AssertionError, pyson.DateTime, 2010, 1, 12, 10, 30,
-                20, 'test', -1, 12, -7, 2, 15, 30, 1)
-        self.assertRaises(AssertionError, pyson.DateTime, 2010, 1, 12, 10, 30,
-                20, 0, 'test', 12, -7, 2, 15, 30, 1)
-        self.assertRaises(AssertionError, pyson.DateTime, 2010, 1, 12, 10, 30,
-                20, 0, -1, 'test', -7, 2, 15, 30, 1)
-        self.assertRaises(AssertionError, pyson.DateTime, 2010, 1, 12, 10, 30,
-                20, 0, -1, 12, 'test', 2, 15, 30, 1)
-        self.assertRaises(AssertionError, pyson.DateTime, 2010, 1, 12, 10, 30,
-                20, 0, -1, 12, -7, 'test', 15, 30, 1)
-        self.assertRaises(AssertionError, pyson.DateTime, 2010, 1, 12, 10, 30,
-                20, 0, -1, 12, -7, 2, 'test', 30, 1)
-        self.assertRaises(AssertionError, pyson.DateTime, 2010, 1, 12, 10, 30,
-                20, 0, -1, 12, -7, 2, 15, 'test', 1)
-        self.assertRaises(AssertionError, pyson.DateTime, 2010, 1, 12, 10, 30,
-                20, 0, -1, 12, -7, 2, 15, 30, 'test')
+        if not sys.flags.optimize:
+            self.assertRaises(
+                AssertionError, pyson.DateTime,
+                'test', 1, 12, 10, 30, 20, 0, -1, 12, -7, 2, 15, 30, 1)
+            self.assertRaises(
+                AssertionError, pyson.DateTime,
+                2010, 'test', 12, 10, 30, 20, 0, -1, 12, -7, 2, 15, 30, 1)
+            self.assertRaises(
+                AssertionError, pyson.DateTime,
+                2010, 1, 'test', 10, 30, 20, 0, -1, 12, -7, 2, 15, 30, 1)
+            self.assertRaises(
+                AssertionError, pyson.DateTime,
+                2010, 1, 12, 'test', 30, 20, 0, -1, 12, -7, 2, 15, 30, 1)
+            self.assertRaises(
+                AssertionError, pyson.DateTime,
+                2010, 1, 12, 10, 'test', 20, 0, -1, 12, -7, 2, 15, 30, 1)
+            self.assertRaises(
+                AssertionError, pyson.DateTime,
+                2010, 1, 12, 10, 30, 'test', 0, -1, 12, -7, 2, 15, 30, 1)
+            self.assertRaises(
+                AssertionError, pyson.DateTime,
+                2010, 1, 12, 10, 30, 20, 'test', -1, 12, -7, 2, 15, 30, 1)
+            self.assertRaises(
+                AssertionError, pyson.DateTime,
+                2010, 1, 12, 10, 30, 20, 0, 'test', 12, -7, 2, 15, 30, 1)
+            self.assertRaises(
+                AssertionError, pyson.DateTime,
+                2010, 1, 12, 10, 30, 20, 0, -1, 'test', -7, 2, 15, 30, 1)
+            self.assertRaises(
+                AssertionError, pyson.DateTime,
+                2010, 1, 12, 10, 30, 20, 0, -1, 12, 'test', 2, 15, 30, 1)
+            self.assertRaises(
+                AssertionError, pyson.DateTime,
+                2010, 1, 12, 10, 30, 20, 0, -1, 12, -7, 'test', 15, 30, 1)
+            self.assertRaises(
+                AssertionError, pyson.DateTime,
+                2010, 1, 12, 10, 30, 20, 0, -1, 12, -7, 2, 'test', 30, 1)
+            self.assertRaises(
+                AssertionError, pyson.DateTime,
+                2010, 1, 12, 10, 30, 20, 0, -1, 12, -7, 2, 15, 'test', 1)
+            self.assertRaises(
+                AssertionError, pyson.DateTime,
+                2010, 1, 12, 10, 30, 20, 0, -1, 12, -7, 2, 15, 30, 'test')
 
         self.assertEqual(pyson.DateTime(2010, 1, 12, 10, 30, 20, 0,
             -1, 12, -7, 2, 15, 30, 1).types(), set([datetime.datetime]))
@@ -543,7 +690,35 @@ class PYSONTestCase(unittest.TestCase):
 
         self.assertEqual(repr(pyson.DateTime(2010, 1, 12, 10, 30, 20, 0,
                     -1, 12, -7, 2, 15, 30, 1)),
-            'DateTime(2010, 1, 12, 10, 30, 20, 0, -1, 12, -7, 2, 15, 30, 1)')
+            ('DateTime(2010, 1, 12, 10, 30, 20, 0, '
+                '-1, 12, -7, 2, 15, 30, 1, None)'))
+
+    def test_DateTime_start(self):
+        "Test DateTime with start"
+        eval = pyson.PYSONEncoder().encode(pyson.DateTime(
+                start=pyson.Eval('datetime')))
+
+        datetime_ = datetime.datetime(2000, 1, 1, 12, 0)
+        self.assertEqual(pyson.PYSONDecoder(
+                {'datetime': datetime_}).decode(eval), datetime_)
+
+    def test_DateTime_start_datetime(self):
+        "Test DateTime with start as date"
+        eval = pyson.PYSONEncoder().encode(pyson.DateTime(
+                start=pyson.Eval('date')))
+
+        date = datetime.date(2000, 1, 1)
+        self.assertEqual(pyson.PYSONDecoder(
+                {'date': date}).decode(eval),
+            datetime.datetime(2000, 1, 1, 0, 0))
+
+    def test_DateTime_start_invalid(self):
+        "Test DateTime with invalid start"
+        eval = pyson.PYSONEncoder().encode(pyson.DateTime(
+                start=pyson.Eval('foo')))
+
+        self.assertIsInstance(pyson.PYSONDecoder(
+                {'foo': 'bar'}).decode(eval), datetime.datetime)
 
     def test_Len(self):
         'Test pyson.Len'
@@ -552,7 +727,8 @@ class PYSONTestCase(unittest.TestCase):
                 'v': [1, 2, 3],
                 })
 
-        self.assertRaises(AssertionError, pyson.Len, object())
+        if not sys.flags.optimize:
+            self.assertRaises(AssertionError, pyson.Len, object())
 
         self.assertEqual(pyson.Len([1, 2, 3]).types(), set([int]))
 
@@ -581,6 +757,7 @@ class PYSONTestCase(unittest.TestCase):
         self.assertEqual(
             pyson.TimeDelta(seconds=10).types(), {datetime.timedelta})
 
+    @unittest.skipIf(sys.flags.optimize, "assert removed by optimization")
     def test_TimeDelta_invalid_type(self):
         "Test pyson.TimeDelta invalid type"
         with self.assertRaises(AssertionError):
@@ -635,6 +812,40 @@ class PYSONTestCase(unittest.TestCase):
             self.assertEqual(decoder.decode(encoder.encode(instance)).pyson(),
                 instance.pyson())
 
+    def test_Eval_dot_notation(self):
+        'Test pyson.Eval with dot notation'
+
+        eval = pyson.PYSONEncoder().encode(pyson.Eval('foo.bar', 0))
+
+        for ctx, result in [
+                ({'foo': {'bar': 1}}, 1),
+                ({'foo': {'foo': 1}}, 0),
+                ({'bar': {'bar': 1}}, 0),
+                ({}, 0),
+                ]:
+            self.assertEqual(pyson.PYSONDecoder(ctx).decode(eval), result)
+
+    def test_Eval_dot_notation_nested(self):
+        'Test pyson.Eval with dot notation with nested dots'
+
+        eval = pyson.PYSONEncoder().encode(pyson.Eval('foo.bar.test', 0))
+
+        for ctx, result in [
+                ({'foo': {'bar': {'test': 1}}}, 1),
+                ({'foo': {'foo': 1}}, 0),
+                ({'bar': {'bar': 1}}, 0),
+                ({}, 0),
+                ]:
+            self.assertEqual(pyson.PYSONDecoder(ctx).decode(eval), result)
+
+    def test_eval_true(self):
+        "Test PYSON.eval JS true"
+        self.assertEqual(eval('true', pyson.CONTEXT), True)
+
+    def test_eval_false(self):
+        "Test PYSON.eval JS false"
+        self.assertEqual(eval('false', pyson.CONTEXT), False)
+
 
 def suite():
     return unittest.TestLoader().loadTestsFromTestCase(PYSONTestCase)
