@@ -9,6 +9,8 @@ import subprocess
 import sys
 import time
 import unittest
+import unittest.mock
+from configparser import ConfigParser
 from functools import reduce
 from functools import wraps
 from itertools import chain
@@ -24,7 +26,7 @@ from trytond.pool import Pool, isregisteredby
 from trytond import backend
 from trytond.model import Workflow, ModelSQL, ModelSingleton, ModelView, fields
 from trytond.model.fields import get_eval_fields, Function
-from trytond.tools import is_instance_method
+from trytond.tools import is_instance_method, file_open
 from trytond.transaction import Transaction
 from trytond.cache import Cache
 from trytond.config import config, parse_uri
@@ -44,29 +46,34 @@ DB_CACHE = os.environ.get('DB_CACHE')
 Pool.test = True
 
 
-def activate_module(name):
+def activate_module(modules, lang='en'):
     '''
-    Activate module for the tested database
+    Activate modules for the tested database
     '''
+    if isinstance(modules, str):
+        modules = [modules]
+    name = '-'.join(modules)
+    if lang != 'en':
+        name += '--' + lang
     if not db_exist(DB_NAME) and restore_db_cache(name):
         return
-    create_db()
+    create_db(lang=lang)
     with Transaction().start(DB_NAME, 1, close=True) as transaction:
         pool = Pool()
         Module = pool.get('ir.module')
 
-        modules = Module.search([
-                ('name', '=', name),
+        records = Module.search([
+                ('name', 'in', modules),
                 ])
-        assert modules, "%s not found" % name
+        assert len(records) == len(modules)
 
-        modules = Module.search([
-                ('name', '=', name),
+        records = Module.search([
+                ('name', 'in', modules),
                 ('state', '!=', 'activated'),
                 ])
 
-        if modules:
-            Module.activate(modules)
+        if records:
+            Module.activate(records)
             transaction.commit()
 
             ActivateUpgrade = pool.get('ir.module.activate_upgrade',
@@ -82,12 +89,11 @@ def activate_module(name):
 def restore_db_cache(name):
     result = False
     if DB_CACHE:
-        backend_name = backend.name()
-        cache_file = _db_cache_file(DB_CACHE, name, backend_name)
+        cache_file = _db_cache_file(DB_CACHE, name)
         if os.path.exists(cache_file):
-            if backend_name == 'sqlite':
+            if backend.name == 'sqlite':
                 result = _sqlite_copy(cache_file, restore=True)
-            elif backend_name == 'postgresql':
+            elif backend.name == 'postgresql':
                 result = _pg_restore(cache_file)
     if result:
         Pool(DB_NAME).init()
@@ -98,23 +104,22 @@ def backup_db_cache(name):
     if DB_CACHE:
         if not os.path.exists(DB_CACHE):
             os.makedirs(DB_CACHE)
-        backend_name = backend.name()
-        cache_file = _db_cache_file(DB_CACHE, name, backend_name)
+        cache_file = _db_cache_file(DB_CACHE, name)
         if not os.path.exists(cache_file):
-            if backend_name == 'sqlite':
+            if backend.name == 'sqlite':
                 _sqlite_copy(cache_file)
-            elif backend_name == 'postgresql':
+            elif backend.name == 'postgresql':
                 _pg_dump(cache_file)
 
 
-def _db_cache_file(path, name, backend_name):
-    return os.path.join(path, '%s-%s.dump' % (name, backend_name))
+def _db_cache_file(path, name):
+    return os.path.join(path, '%s-%s.dump' % (name, backend.name))
 
 
 def _sqlite_copy(file_, restore=False):
     import sqlite3 as sqlite
 
-    with Transaction().start(DB_NAME, 0, _nocache=True) as transaction, \
+    with Transaction().start(DB_NAME, 0) as transaction, \
             sqlite.connect(file_) as conn2:
         conn1 = transaction.connection
         # sqlitebck does not work with pysqlite2
@@ -150,8 +155,7 @@ def _pg_options():
 
 def _pg_restore(cache_file):
     with Transaction().start(
-            None, 0, close=True, autocommit=True, _nocache=True) \
-            as transaction:
+            None, 0, close=True, autocommit=True) as transaction:
         transaction.database.create(transaction.connection, DB_NAME)
     cmd = ['pg_restore', '-d', DB_NAME]
     options, env = _pg_options()
@@ -161,9 +165,9 @@ def _pg_restore(cache_file):
         return not subprocess.call(cmd, env=env)
     except OSError:
         cache_name, _ = os.path.splitext(os.path.basename(cache_file))
+        cache_name = backend.TableHandler.convert_name(cache_name)
         with Transaction().start(
-                None, 0, close=True, autocommit=True, _nocache=True) \
-                as transaction:
+                None, 0, close=True, autocommit=True) as transaction:
             transaction.database.drop(transaction.connection, DB_NAME)
             transaction.database.create(
                 transaction.connection, DB_NAME, cache_name)
@@ -179,11 +183,11 @@ def _pg_dump(cache_file):
         return not subprocess.call(cmd, env=env)
     except OSError:
         cache_name, _ = os.path.splitext(os.path.basename(cache_file))
+        cache_name = backend.TableHandler.convert_name(cache_name)
         # Ensure any connection is left open
-        backend.get('Database')(DB_NAME).close()
+        backend.Database(DB_NAME).close()
         with Transaction().start(
-                None, 0, close=True, autocommit=True, _nocache=True) \
-                as transaction:
+                None, 0, close=True, autocommit=True) as transaction:
             transaction.database.create(
                 transaction.connection, cache_name, DB_NAME)
         open(cache_file, 'a').close()
@@ -196,10 +200,12 @@ def with_transaction(user=1, context=None):
         def wrapper(*args, **kwargs):
             transaction = Transaction()
             with transaction.start(DB_NAME, user, context=context):
-                result = func(*args, **kwargs)
-                transaction.rollback()
-                # Drop the cache as the transaction is rollbacked
-                Cache.drop(DB_NAME)
+                try:
+                    result = func(*args, **kwargs)
+                finally:
+                    transaction.rollback()
+                    # Drop the cache as the transaction is rollbacked
+                    Cache.drop(DB_NAME)
                 return result
         return wrapper
     return decorator
@@ -208,11 +214,16 @@ def with_transaction(user=1, context=None):
 class ModuleTestCase(unittest.TestCase):
     'Trytond Test Case'
     module = None
+    extras = None
+    language = 'en'
 
     @classmethod
     def setUpClass(cls):
         drop_db()
-        activate_module(cls.module)
+        modules = [cls.module]
+        if cls.extras:
+            modules.extend(cls.extras)
+        activate_module(modules, lang=cls.language)
         super(ModuleTestCase, cls).setUpClass()
 
     @classmethod
@@ -229,13 +240,29 @@ class ModuleTestCase(unittest.TestCase):
             # as there is a fallback to id
             if model._rec_name == 'name':
                 continue
-            assert model._rec_name in model._fields, (
-                'Wrong _rec_name "%s" for %s'
-                % (model._rec_name, mname))
+            self.assertIn(model._rec_name, model._fields.keys(),
+                msg='Wrong _rec_name "%s" for %s' % (model._rec_name, mname))
             field = model._fields[model._rec_name]
-            assert field._type in {'char', 'text'}, (
-                "Wrong '%s' type for _rec_name of %s'"
-                % (field._type, mname))
+            self.assertIn(field._type, {'char', 'text'},
+                msg="Wrong '%s' type for _rec_name of %s'" % (
+                    field._type, mname))
+
+    @with_transaction()
+    def test_model__access__(self):
+        "Test existing model __access__"
+        pool = Pool()
+        for mname, Model in pool.iterobject():
+            if not isregisteredby(Model, self.module):
+                continue
+            for field_name in Model.__access__:
+                self.assertIn(field_name, Model._fields.keys(),
+                    msg="Wrong __access__ '%s' for %s" % (field_name, mname))
+                field = Model._fields[field_name]
+                Target = field.get_target()
+                self.assertTrue(
+                    Target,
+                    msg='Missing target for __access__ "%s" of %s' % (
+                        field_name, mname))
 
     @with_transaction()
     def test_view(self):
@@ -247,6 +274,13 @@ class ModuleTestCase(unittest.TestCase):
                 ('model', '!=', ''),
                 ])
         for view in views:
+            if not view.inherit or view.inherit.model == view.model:
+                self.assertTrue(view.arch,
+                    msg='missing architecture for view "%(name)s" '
+                    'of model "%(model)s"' % {
+                        'name': view.name or str(view.id),
+                        'model': view.model,
+                        })
             if view.inherit and view.inherit.model == view.model:
                 view_id = view.inherit.id
             else:
@@ -254,39 +288,48 @@ class ModuleTestCase(unittest.TestCase):
             model = view.model
             Model = pool.get(model)
             res = Model.fields_view_get(view_id)
-            assert res['model'] == model
+            self.assertEqual(res['model'], model)
             tree = etree.fromstring(res['arch'])
 
             validator = etree.RelaxNG(etree=View.get_rng(res['type']))
-            # Don't use assert_ because 2to3 convert to assertTrue
             validator.assertValid(tree)
 
             tree_root = tree.getroottree().getroot()
 
             for element in tree_root.iter():
                 if element.tag in ('field', 'label', 'separator', 'group'):
-                    for attr in ('name', 'icon'):
+                    for attr in ['name', 'icon', 'symbol']:
                         field = element.get(attr)
                         if field:
-                            assert field in res['fields'], (
-                                'Missing field: %s' % field)
+                            self.assertIn(field, res['fields'].keys(),
+                                msg='Missing field: %s in %s' % (
+                                    field, Model.__name__))
                 if element.tag == 'button':
                     button_name = element.get('name')
-                    assert button_name in Model._buttons, (
-                        "Button '%s' is not in %s._buttons"
+                    self.assertIn(button_name, Model._buttons.keys(),
+                        msg="Button '%s' is not in %s._buttons"
                         % (button_name, Model.__name__))
 
+    @with_transaction()
+    def test_icon(self):
+        "Test icons of the module"
+        pool = Pool()
+        Icon = pool.get('ir.ui.icon')
+        icons = Icon.search([('module', '=', self.module)])
+        for icon in icons:
+            self.assertTrue(icon.icon)
+
     @with_transaction()
     def test_rpc_callable(self):
         'Test that RPC methods are callable'
         for _, model in Pool().iterobject():
             for method_name in model.__rpc__:
-                assert callable(getattr(model, method_name, None)), (
-                    "'%s' is not callable on '%s'"
+                self.assertTrue(callable(getattr(model, method_name, None)),
+                    msg="'%s' is not callable on '%s'"
                     % (method_name, model.__name__))
 
     @with_transaction()
-    def test_depends(self):
+    def test_missing_depends(self):
         'Test for missing depends'
         for mname, model in Pool().iterobject():
             if not isregisteredby(model, self.module):
@@ -304,22 +347,63 @@ class ModuleTestCase(unittest.TestCase):
                 fields.discard('context')
                 fields.discard('_user')
                 depends = set(field.depends)
-                assert fields <= depends, (
-                    'Missing depends %s in "%s"."%s"' % (
+                self.assertLessEqual(fields, depends,
+                    msg='Missing depends %s in "%s"."%s"' % (
                         list(fields - depends), mname, fname))
-                assert depends <= set(model._fields), (
-                    'Unknown depends %s in "%s"."%s"' % (
+                self.assertLessEqual(depends, set(model._fields),
+                    msg='Unknown depends %s in "%s"."%s"' % (
                         list(depends - set(model._fields)), mname, fname))
             if issubclass(model, ModelView):
                 for bname, button in model._buttons.items():
                     depends = set(button.get('depends', []))
-                    assert depends <= set(model._fields), (
-                        'Unknown depends %s in button "%s"."%s"' % (
+                    self.assertLessEqual(depends, set(model._fields),
+                        msg='Unknown depends %s in button "%s"."%s"' % (
                             list(depends - set(model._fields)), mname, bname))
 
     @with_transaction()
-    def test_depends_parent(self):
-        "Test depends on _parent_ contains also the parent relation"
+    def test_depends(self):
+        "Test depends"
+        def test_missing_relation(depend, depends, qualname):
+            prefix = []
+            for d in depend.split('.'):
+                if d.startswith('_parent_'):
+                    relation = '.'.join(
+                        prefix + [d[len('_parent_'):]])
+                    self.assertIn(relation, depends,
+                        msg='Missing "%s" in %s' % (relation, qualname))
+                prefix.append(d)
+
+        def test_parent_empty(depend, qualname):
+            if depend.startswith('_parent_'):
+                self.assertIn('.', depend,
+                    msg='Invalid empty "%s" in %s' % (depend, qualname))
+
+        def test_missing_parent(model, depend, depends, qualname):
+            dfield = model._fields.get(depend)
+            parent_depends = {d.split('.', 1)[0] for d in depends}
+            if dfield and dfield._type == 'many2one':
+                target = dfield.get_target()
+                for tfield in target._fields.values():
+                    if (tfield._type == 'one2many'
+                            and tfield.model_name == mname
+                            and tfield.field == depend):
+                        self.assertIn('_parent_%s' % depend, parent_depends,
+                            msg='Missing "_parent_%s" in %s' % (
+                                depend, qualname))
+
+        def test_depend_exists(model, depend, qualname):
+            try:
+                depend, nested = depend.split('.', 1)
+            except ValueError:
+                nested = None
+            if depend.startswith('_parent_'):
+                depend = depend[len('_parent_'):]
+            self.assertIsInstance(getattr(model, depend, None), fields.Field,
+                msg='Unknown "%s" in %s' % (depend, qualname))
+            if nested:
+                target = getattr(model, depend).get_target()
+                test_depend_exists(target, nested, qualname)
+
         for mname, model in Pool().iterobject():
             if not isregisteredby(model, self.module):
                 continue
@@ -327,16 +411,14 @@ class ModuleTestCase(unittest.TestCase):
                 for attribute in ['depends', 'on_change', 'on_change_with',
                         'selection_change_with', 'autocomplete']:
                     depends = getattr(field, attribute, [])
+                    qualname = '"%s"."%s"."%s"' % (mname, fname, attribute)
                     for depend in depends:
-                        prefix = []
-                        for d in depend.split('.'):
-                            if d.startswith('_parent_'):
-                                relation = '.'.join(
-                                    prefix + [d[len('_parent_'):]])
-                                assert relation in depends, (
-                                    'Missing "%s" in "%s"."%s"."%s"' % (
-                                        relation, mname, fname, attribute))
-                            prefix.append(d)
+                        test_depend_exists(model, depend, qualname)
+                        test_missing_relation(depend, depends, qualname)
+                        test_parent_empty(depend, qualname)
+                        if attribute != 'depends':
+                            test_missing_parent(
+                                model, depend, depends, qualname)
 
     @with_transaction()
     def test_field_methods(self):
@@ -363,8 +445,8 @@ class ModuleTestCase(unittest.TestCase):
                         if attr.startswith(prefix)]
                     if not fnames:
                         continue
-                    assert any(f in model._fields for f in fnames), (
-                        'Field method "%s"."%s" for unknown field' % (
+                    self.assertTrue(any(f in model._fields for f in fnames),
+                        msg='Field method "%s"."%s" for unknown field' % (
                             mname, attr))
 
                     if attr.startswith('default_'):
@@ -388,6 +470,49 @@ class ModuleTestCase(unittest.TestCase):
                         record = model()
                         getattr(record, attr)()
 
+    @with_transaction()
+    def test_field_relation_target(self):
+        "Test field relation and target"
+        pool = Pool()
+        for mname, model in pool.iterobject():
+            if not isregisteredby(model, self.module):
+                continue
+            for fname, field in model._fields.items():
+                if isinstance(field, fields.One2Many):
+                    Relation = field.get_target()
+                    rfield = field.field
+                elif isinstance(field, fields.Many2Many):
+                    Relation = field.get_relation()
+                    rfield = field.origin
+                else:
+                    continue
+                if rfield:
+                    self.assertIn(rfield, Relation._fields.keys(),
+                        msg=('Missing relation field "%s" on "%s" '
+                            'for "%s"."%s"') % (
+                            rfield, Relation.__name__, mname, fname))
+                    reverse_field = Relation._fields[rfield]
+                    self.assertIn(
+                        reverse_field._type, [
+                            'reference', 'many2one', 'one2one'],
+                        msg=('Wrong type for relation field "%s" on "%s" '
+                            'for "%s"."%s"') % (
+                            rfield, Relation.__name__, mname, fname))
+                    if (reverse_field._type == 'many2one'
+                            and issubclass(model, ModelSQL)
+                            # Do not test table_query models
+                            # as they can manipulate their id
+                            and not callable(model.table_query)):
+                        self.assertEqual(
+                            reverse_field.model_name, model.__name__,
+                            msg=('Wrong model for relation field "%s" on "%s" '
+                                'for "%s"."%s"') % (
+                                rfield, Relation.__name__, mname, fname))
+                Target = field.get_target()
+                self.assertTrue(
+                    Target,
+                    msg='Missing target for "%s"."%s"' % (mname, fname))
+
     @with_transaction()
     def test_menu_action(self):
         'Test that menu actions are accessible to menu\'s group'
@@ -409,8 +534,8 @@ class ModuleTestCase(unittest.TestCase):
                     if k.keyword == 'tree_open'))
             if not actions_groups:
                 continue
-            assert menu_groups <= actions_groups, (
-                'Menu "%(menu_xml_id)s" actions are not accessible to '
+            self.assertLessEqual(menu_groups, actions_groups,
+                msg='Menu "%(menu_xml_id)s" actions are not accessible to '
                 '%(groups)s' % {
                     'menu_xml_id': module_menu.fs_id,
                     'groups': ','.join(g.name
@@ -422,15 +547,27 @@ class ModuleTestCase(unittest.TestCase):
         'Test missing default model access'
         pool = Pool()
         Access = pool.get('ir.model.access')
-        no_groups = {a.model.name for a in Access.search([
+        no_groups = {a.model.model for a in Access.search([
                     ('group', '=', None),
                     ])}
-        with_groups = {a.model.name for a in Access.search([
+
+        def has_access(Model, models):
+            if Model.__name__ in models:
+                return True
+            for field_name in Model.__access__:
+                Target = Model._fields[field_name].get_target()
+                if has_access(Target, models):
+                    return True
+        for mname, Model in pool.iterobject():
+            if has_access(Model, no_groups):
+                no_groups.add(mname)
+
+        with_groups = {a.model.model for a in Access.search([
                     ('group', '!=', None),
                     ])}
 
-        assert no_groups >= with_groups, (
-            'Model "%(models)s" are missing a default access' % {
+        self.assertGreaterEqual(no_groups, with_groups,
+            msg='Model "%(models)s" are missing a default access' % {
                 'models': list(with_groups - no_groups),
                 })
 
@@ -452,9 +589,9 @@ class ModuleTestCase(unittest.TestCase):
                 values = getattr(model, field.selection)()
             states = set(dict(values))
             transition_states = set(chain(*model._transitions))
-            assert transition_states <= states, (
-                ('Unknown transition states "%(states)s" '
-                    'in model "%(model)s". ') % {
+            self.assertLessEqual(transition_states, states,
+                msg='Unknown transition states "%(states)s" '
+                'in model "%(model)s". ' % {
                     'states': list(transition_states - states),
                     'model': model.__name__,
                     })
@@ -466,7 +603,8 @@ class ModuleTestCase(unittest.TestCase):
             if not isregisteredby(wizard, self.module, type_='wizard'):
                 continue
             session_id, start_state, _ = wizard.create()
-            assert start_state in wizard.states, ('Unknown start state '
+            self.assertIn(start_state, wizard.states.keys(),
+                msg='Unknown start state '
                 '"%(state)s" on wizard "%(wizard)s"' % {
                     'state': start_state,
                     'wizard': wizard_name,
@@ -475,8 +613,26 @@ class ModuleTestCase(unittest.TestCase):
             for state_name, state in wizard_instance.states.items():
                 if isinstance(state, StateView):
                     # Don't test defaults as they may depend on context
-                    state.get_view(wizard_instance, state_name)
-                    state.get_buttons(wizard_instance, state_name)
+                    view = state.get_view(wizard_instance, state_name)
+                    self.assertEqual(
+                        view.get('type'), 'form',
+                        msg='Wrong view type for "%(state)s" '
+                        'on wizard "%(wizard)s"' % {
+                            'state': state_name,
+                            'wizard': wizard_name,
+                            })
+                    for button in state.get_buttons(
+                            wizard_instance, state_name):
+                        if button['state'] == wizard.end_state:
+                            continue
+                        self.assertIn(
+                            button['state'],
+                            wizard_instance.states.keys(),
+                            msg='Unknown button state from "%(state)s" '
+                            'on wizard "%(wizard)s' % {
+                                'state': state_name,
+                                'wizard': wizard_name,
+                                })
                 if isinstance(state, StateAction):
                     state.get_action()
 
@@ -498,8 +654,8 @@ class ModuleTestCase(unittest.TestCase):
                     else:
                         record = model()
                         selection_values = sel_func(record)
-                assert all(len(v) == 2 for v in selection_values), (
-                    'Invalid selection values "%(values)s" on field '
+                self.assertTrue(all(len(v) == 2 for v in selection_values),
+                    msg='Invalid selection values "%(values)s" on field '
                     '"%(field)s" of model "%(model)s"' % {
                         'values': selection_values,
                         'field': field_name,
@@ -518,9 +674,9 @@ class ModuleTestCase(unittest.TestCase):
                 for func_name in [field.getter, field.setter, field.searcher]:
                     if not func_name:
                         continue
-                    assert getattr(model, func_name, None), (
-                        "Missing method '%(func_name)s' "
-                        "on model '%(model)s' for field '%(field)s" % {
+                    self.assertTrue(getattr(model, func_name, None),
+                        msg="Missing method '%(func_name)s' "
+                        "on model '%(model)s' for field '%(field)s'" % {
                             'func_name': func_name,
                             'model': model.__name__,
                             'field': field_name,
@@ -581,8 +737,8 @@ class ModuleTestCase(unittest.TestCase):
             mro = inspect.getmro(model)
             singleton_index = mro.index(ModelSingleton)
             sql_index = mro.index(ModelSQL)
-            assert singleton_index < sql_index, (
-                "ModelSingleton must appear before ModelSQL in the parent "
+            self.assertLess(singleton_index, sql_index,
+                msg="ModelSingleton must appear before ModelSQL in the parent "
                 "classes of '%s'." % mname)
 
     @with_transaction()
@@ -599,38 +755,81 @@ class ModuleTestCase(unittest.TestCase):
                         ('model.model', '=', model.__name__),
                         ])}
             buttons = set(model._buttons)
-            assert ir_buttons >= buttons, (
-                'The buttons "%(buttons)s" of Model "%(model)s" are not '
+            self.assertGreaterEqual(ir_buttons, buttons,
+                msg='The buttons "%(buttons)s" of Model "%(model)s" are not '
                 'registered in ir.model.button.' % {
                     'buttons': list(buttons - ir_buttons),
                     'model': model.__name__,
                     })
 
+    @with_transaction()
+    def test_buttons_states(self):
+        "Test the states of buttons"
+        pool = Pool()
+        keys = {'readonly', 'invisible', 'icon', 'pre_validate', 'depends'}
+        for mname, model in pool.iterobject():
+            if not isregisteredby(model, self.module):
+                continue
+            if not issubclass(model, ModelView):
+                continue
+            for button, states in model._buttons.items():
+                self.assertTrue(set(states).issubset(keys),
+                    msg='The button "%(button)s" of Model "%(model)s" has '
+                    'extra keys "%(keys)s".' % {
+                        'button': button,
+                        'model': mname,
+                        'keys': set(states) - keys,
+                        })
+
+    @with_transaction()
+    def test_xml_files(self):
+        "Test validity of the xml files of the module"
+        config = ConfigParser()
+        with file_open('%s/tryton.cfg' % self.module,
+                subdir='modules', mode='r', encoding='utf-8') as fp:
+            config.read_file(fp)
+        if not config.has_option('tryton', 'xml'):
+            return
+        with file_open('tryton.rng', subdir='', mode='rb') as fp:
+            rng = etree.parse(fp)
+        validator = etree.RelaxNG(etree=rng)
+        for xml_file in filter(None, config.get('tryton', 'xml').splitlines()):
+            with file_open('%s/%s' % (self.module, xml_file),
+                    subdir='modules', mode='rb') as fp:
+                tree = etree.parse(fp)
+            validator.assertValid(tree)
+
 
 def db_exist(name=DB_NAME):
-    Database = backend.get('Database')
-    database = Database().connect()
+    database = backend.Database().connect()
     return name in database.list()
 
 
 def create_db(name=DB_NAME, lang='en'):
-    Database = backend.get('Database')
     if not db_exist(name):
-        with Transaction().start(
-                None, 0, close=True, autocommit=True, _nocache=True) \
-                as transaction:
-            transaction.database.create(transaction.connection, name)
+        database = backend.Database()
+        database.connect()
+        connection = database.get_connection(autocommit=True)
+        try:
+            database.create(connection, name)
+        finally:
+            database.put_connection(connection, True)
 
-        with Transaction().start(name, 0, _nocache=True) as transaction,\
-                transaction.connection.cursor() as cursor:
-            Database(name).init()
-            ir_configuration = Table('ir_configuration')
-            cursor.execute(*ir_configuration.insert(
-                    [ir_configuration.language], [[lang]]))
+        database = backend.Database(name)
+        connection = database.get_connection()
+        try:
+            with connection.cursor() as cursor:
+                database.init()
+                ir_configuration = Table('ir_configuration')
+                cursor.execute(*ir_configuration.insert(
+                        [ir_configuration.language], [[lang]]))
+            connection.commit()
+        finally:
+            database.put_connection(connection)
 
         pool = Pool(name)
         pool.init(update=['res', 'ir'], lang=[lang])
-        with Transaction().start(name, 0) as transaction:
+        with Transaction().start(name, 0):
             User = pool.get('res.user')
             Lang = pool.get('ir.lang')
             language, = Lang.search([('code', '=', lang)])
@@ -644,15 +843,49 @@ def create_db(name=DB_NAME, lang='en'):
             Module.update_list()
 
 
+class ExtensionTestCase(unittest.TestCase):
+    extension = None
+
+    @classmethod
+    def setUpClass(cls):
+        super().setUpClass()
+        cls._activate_extension()
+
+    @classmethod
+    def tearDownClass(cls):
+        super().tearDownClass()
+        cls._deactivate_extension()
+
+    @classmethod
+    @with_transaction()
+    def _activate_extension(cls):
+        connection = Transaction().connection
+        cursor = connection.cursor()
+        cursor.execute('CREATE EXTENSION "%s"' % cls.extension)
+        connection.commit()
+        cls._clear_cache()
+
+    @classmethod
+    @with_transaction()
+    def _deactivate_extension(cls):
+        connection = Transaction().connection
+        cursor = connection.cursor()
+        cursor.execute('DROP EXTENSION "%s"' % cls.extension)
+        connection.commit()
+        cls._clear_cache()
+
+    @classmethod
+    def _clear_cache(cls):
+        backend.Database._has_proc.clear()
+
+
 def drop_db(name=DB_NAME):
     if db_exist(name):
-        Database = backend.get('Database')
-        database = Database(name)
+        database = backend.Database(name)
         database.close()
 
         with Transaction().start(
-                None, 0, close=True, autocommit=True, _nocache=True) \
-                as transaction:
+                None, 0, close=True, autocommit=True) as transaction:
             database.drop(transaction.connection, name)
             Pool.stop(name)
             Cache.drop(name)
@@ -669,6 +902,7 @@ def doctest_setup(test):
 
 
 def doctest_teardown(test):
+    unittest.mock.patch.stopall()
     return drop_db()
 
 
@@ -677,12 +911,11 @@ doctest_checker = doctest.OutputChecker()
 
 class TestSuite(unittest.TestSuite):
     def run(self, *args, **kwargs):
-        DatabaseOperationalError = backend.get('DatabaseOperationalError')
         while True:
             try:
                 exist = db_exist()
                 break
-            except DatabaseOperationalError as err:
+            except backend.DatabaseOperationalError as err:
                 # Retry on connection error
                 sys.stderr.write(str(err))
                 time.sleep(1)
