@@ -4,16 +4,20 @@
 """
 Miscelleanous tools used by tryton
 """
-import os
-import sys
-from array import array
-from itertools import islice
-import types
-import io
-import warnings
 import importlib
+import io
+import os
+import re
+import sys
+import types
+import unicodedata
+import warnings
+from array import array
+from functools import wraps
+from itertools import islice
 
 from sql import Literal
+from sql.conditionals import Case
 from sql.operators import Or
 
 from trytond.const import OPERATORS
@@ -81,76 +85,13 @@ def get_smtp_server():
     :return: A SMTP instance. The quit() method must be call when all
     the calls to sendmail() have been made.
     """
-    from ..sendmail import get_smtp_server
+    from trytond.sendmail import get_smtp_server
     warnings.warn(
         'get_smtp_server is deprecated use trytond.sendmail',
         DeprecationWarning)
     return get_smtp_server()
 
 
-def memoize(maxsize):
-    """
-    Decorator to 'memoize' a function - caching its results with a
-    near LRU implementation.
-
-    The cache keeps a list of keys logicaly separated in 4 segment :
-
-    segment 1 | ...        | segment4
-    [k,k,k,k,k,k,k, .. ,k,k,k,k,k,k,k]
-
-    For each segment there is a pointer that loops on it.  When a key
-    is accessed from the cache it is promoted to the first segment (at
-    the pointer place of segment one), the key under the pointer is
-    moved to the next segment, the pointer is then incremented and so
-    on. A key that is removed from the last segment is removed from
-    the cache.
-
-    :param: maxsize the size of the cache (must be greater than or
-    equal to 4)
-    """
-    assert maxsize >= 4, "Memoize cannot work if maxsize is less than 4"
-
-    def wrap(fct):
-        cache = {}
-        keys = [None for i in range(maxsize)]
-        seg_size = maxsize // 4
-
-        pointers = [i * seg_size for i in range(4)]
-        max_pointers = [(i + 1) * seg_size for i in range(3)] + [maxsize]
-
-        def wrapper(*args):
-            key = repr(args)
-            res = cache.get(key)
-            if res:
-                pos, res = res
-                keys[pos] = None
-            else:
-                res = fct(*args)
-
-            value = res
-            for segment, pointer in enumerate(pointers):
-                newkey = keys[pointer]
-                keys[pointer] = key
-                cache[key] = (pointer, value)
-
-                pointers[segment] = pointer + 1
-                if pointers[segment] == max_pointers[segment]:
-                    pointers[segment] = segment * seg_size
-
-                if newkey is None:
-                    break
-                segment, value = cache.pop(newkey)
-                key = newkey
-
-            return res
-
-        wrapper.__doc__ = fct.__doc__
-        wrapper.__name__ = fct.__name__
-
-        return wrapper
-    return wrap
-
-
 def reduce_ids(field, ids):
     '''
     Return a small SQL expression for the list of ids and the sql column
@@ -201,10 +142,10 @@ def reduce_domain(domain):
         domain = domain[1:]
     result = [operator]
     for arg in domain:
-        if (isinstance(arg, tuple) or
-                (isinstance(arg, list) and
-                    len(arg) > 2 and
-                    arg[1] in OPERATORS)):
+        if (isinstance(arg, tuple)
+                or (isinstance(arg, list)
+                    and len(arg) > 2
+                    and arg[1] in OPERATORS)):
             # clause
             result.append(arg)
         elif isinstance(arg, list) and arg:
@@ -225,6 +166,7 @@ def grouped_slice(records, count=None):
     from trytond.transaction import Transaction
     if count is None:
         count = Transaction().database.IN_MAX
+    count = max(1, count)
     for i in range(0, len(records), count):
         yield islice(records, i, i + count)
 
@@ -248,3 +190,85 @@ def resolve(name):
         except AttributeError:
             found = importlib.import_module(used)
     return found
+
+
+def strip_wildcard(string, wildcard='%', escape='\\'):
+    "Strip starting and ending wildcard from string"
+    string = lstrip_wildcard(string, wildcard)
+    return rstrip_wildcard(string, wildcard, escape)
+
+
+def lstrip_wildcard(string, wildcard='%'):
+    "Strip starting wildcard from string"
+    if not string:
+        return string
+    return string.lstrip(wildcard)
+
+
+def rstrip_wildcard(string, wildcard='%', escape='\\'):
+    "Strip ending wildcard from string"
+    if not string:
+        return string
+    new_string = string.rstrip(wildcard)
+    if new_string[-1] == escape:
+        return string
+    return new_string
+
+
+def escape_wildcard(string, wildcards='%_', escape='\\'):
+    for wildcard in escape + wildcards:
+        string = string.replace(wildcard, escape + wildcard)
+    return string
+
+
+def unescape_wildcard(string, wildcards='%_', escape='\\'):
+    for wildcard in wildcards + escape:
+        string = string.replace(escape + wildcard, wildcard)
+    return string
+
+
+def is_full_text(value, escape='\\'):
+    escaped = value.strip('%')
+    escaped = escaped.replace(escape + '%', '').replace(escape + '_', '')
+    if '%' in escaped or '_' in escaped:
+        return False
+    return value.startswith('%') == value.endswith('%')
+
+
+_slugify_strip_re = re.compile(r'[^\w\s-]')
+_slugify_hyphenate_re = re.compile(r'[-\s]+')
+
+
+def slugify(value, hyphenate='-'):
+    if not isinstance(value, str):
+        value = str(value)
+    value = unicodedata.normalize('NFKD', value)
+    value = str(_slugify_strip_re.sub('', value).strip())
+    return _slugify_hyphenate_re.sub(hyphenate, value)
+
+
+def sortable_values(func):
+    "Decorator that makes list of couple values sortable"
+    @wraps(func)
+    def wrapper(*args, **kwargs):
+        result = list(func(*args, **kwargs))
+        for i, (name, value) in enumerate(list(result)):
+            result[i] = (name, value is None, value)
+        return result
+    return wrapper
+
+
+def sql_pairing(x, y):
+    """Return SQL expression to pair x and y
+    Pairing function from http://szudzik.com/ElegantPairing.pdf"""
+    return Case(
+        (x < y, (y * y) + x),
+        else_=(x * x) + x + y)
+
+
+def firstline(text):
+    "Returns first non-empty line"
+    try:
+        return next((x for x in text.splitlines() if x.strip()))
+    except StopIteration:
+        return ''
