@@ -1,12 +1,14 @@
 # This file is part of Tryton.  The COPYRIGHT file at the top level of
 # this repository contains the full copyright notices and license terms.
 import logging
+import time
+from collections import defaultdict
 from threading import local
 from sql import Flavor
 
-from trytond import backend
 from trytond.config import config
 
+_cache_model = config.getint('cache', 'model')
 logger = logging.getLogger(__name__)
 
 
@@ -52,7 +54,9 @@ class Transaction(object):
     create_records = None
     delete_records = None
     delete = None  # TODO check to merge with delete_records
+    trigger_records = None
     timestamp = None
+    started_at = None
 
     def __new__(cls, new=False):
         transactions = cls._local.transactions
@@ -65,6 +69,13 @@ class Transaction(object):
             instance = transactions[-1]
         return instance
 
+    @staticmethod
+    def monotonic_time():
+        try:
+            return time.monotonic_ns()
+        except AttributeError:
+            return time.monotonic()
+
     @property
     def tasks(self):
         return self._local.tasks
@@ -74,24 +85,27 @@ class Transaction(object):
         keys = tuple(((key, self.context[key])
                 for key in sorted(self.cache_keys)
                 if key in self.context))
-        return self.cache.setdefault((self.user, keys),
-            LRUDict(config.getint('cache', 'model')))
+        return self.cache.setdefault((self.user, keys), LRUDict(_cache_model))
 
     def start(self, database_name, user, readonly=False, context=None,
-            close=False, autocommit=False, _nocache=False):
+            close=False, autocommit=False):
         '''
         Start transaction
         '''
-        Database = backend.get('Database')
+        from trytond import backend
         assert self.user is None
         assert self.database is None
         assert self.close is None
         assert self.context is None
+        # Compute started_at before connect to ensure
+        # it is strictly before all transactions started after
+        # but it may be also before transactions started before
+        self.started_at = self.monotonic_time()
         if not database_name:
-            database = Database().connect()
+            database = backend.Database().connect()
         else:
-            database = Database(database_name).connect()
-        Flavor.set(Database.flavor)
+            database = backend.Database(database_name).connect()
+        Flavor.set(backend.Database.flavor)
         self.connection = database.get_connection(readonly=readonly,
             autocommit=autocommit)
         self.user = user
@@ -102,14 +116,14 @@ class Transaction(object):
         self.create_records = {}
         self.delete_records = {}
         self.delete = {}
+        self.trigger_records = defaultdict(set)
         self.timestamp = {}
         self.counter = 0
         self._datamanagers = []
-        self._nocache = _nocache
-        if not _nocache:
+        if database_name:
             from trytond.cache import Cache
             try:
-                Cache.clean(database.name)
+                Cache.sync(self)
             except BaseException:
                 self.stop(False)
                 raise
@@ -146,6 +160,7 @@ class Transaction(object):
                     self.create_records = None
                     self.delete_records = None
                     self.delete = None
+                    self.trigger_records = None
                     self.timestamp = None
                     self._datamanagers = []
 
@@ -191,14 +206,14 @@ class Transaction(object):
         self._local.transactions.append(transaction)
         return transaction
 
-    def new_transaction(self, autocommit=False, readonly=False,
-            _nocache=False):
+    def new_transaction(self, autocommit=False, readonly=False):
         transaction = Transaction(new=True)
         return transaction.start(self.database.name, self.user,
             context=self.context, close=self.close, readonly=readonly,
-            autocommit=autocommit, _nocache=_nocache)
+            autocommit=autocommit)
 
     def commit(self):
+        from trytond.cache import Cache
         try:
             if self._datamanagers:
                 for datamanager in self._datamanagers:
@@ -207,31 +222,29 @@ class Transaction(object):
                     datamanager.commit(self)
                 for datamanager in self._datamanagers:
                     datamanager.tpc_vote(self)
+            self.started_at = self.monotonic_time()
+            Cache.commit(self)
             self.connection.commit()
-        except:
+        except Exception:
             self.rollback()
             raise
         else:
             try:
                 for datamanager in self._datamanagers:
                     datamanager.tpc_finish(self)
-            except:
+            except Exception:
                 logger.critical('A datamanager raised an exception in'
                     ' tpc_finish, the data might be inconsistant',
                     exc_info=True)
-        if not self._nocache:
-            from trytond.cache import Cache
-            Cache.resets(self.database.name)
 
     def rollback(self):
+        from trytond.cache import Cache
         for cache in self.cache.values():
             cache.clear()
         for datamanager in self._datamanagers:
             datamanager.tpc_abort(self)
+        Cache.rollback(self)
         self.connection.rollback()
-        if not self._nocache:
-            from trytond.cache import Cache
-            Cache.resets(self.database.name)
 
     def join(self, datamanager):
         try:
