@@ -13,17 +13,16 @@ from sql import Flavor
 from trytond import backend
 from trytond.config import config
 from trytond.pool import Pool
+from trytond.status import processing
 from trytond.transaction import Transaction
 
 __all__ = ['work']
 logger = logging.getLogger(__name__)
-Database = backend.get('Database')
-DatabaseOperationalError = backend.get('DatabaseOperationalError')
 
 
 class Queue(object):
     def __init__(self, pool, mpool):
-        self.database = Database(pool.database_name).connect()
+        self.database = backend.Database(pool.database_name).connect()
         self.connection = self.database.get_connection(autocommit=True)
         self.pool = pool
         self.mpool = mpool
@@ -46,7 +45,7 @@ class TaskList(list):
 
 
 def work(options):
-    Flavor.set(Database.flavor)
+    Flavor.set(backend.Database.flavor)
     if not config.getboolean('queue', 'worker', default=False):
         return
     try:
@@ -100,20 +99,33 @@ def run_task(pool, task_id):
     if not isinstance(pool, Pool):
         pool = Pool(pool)
     Queue = pool.get('ir.queue')
-    logger.info('task "%d" started', task_id)
+    name = '<Task %s@%s>' % (task_id, pool.database_name)
+    logger.info('%s started', name)
+    retry = config.getint('database', 'retry')
     try:
-        for count in range(config.getint('database', 'retry'), -1, -1):
+        for count in range(retry, -1, -1):
+            if count != retry:
+                time.sleep(0.02 * (retry - count))
             with Transaction().start(pool.database_name, 0) as transaction:
                 try:
-                    Queue(task_id).run()
+                    try:
+                        task, = Queue.search([('id', '=', task_id)])
+                    except ValueError:
+                        # the task was rollbacked, nothing to do
+                        break
+                    with processing(name):
+                        task.run()
                     break
-                except DatabaseOperationalError:
+                except backend.DatabaseOperationalError:
                     if count:
                         transaction.rollback()
                         continue
                     raise
-        logger.info('task "%d" done', task_id)
-    except DatabaseOperationalError:
+        logger.info('%s done', name)
+    except backend.DatabaseOperationalError:
+        logger.info('%s failed, retrying', name, exc_info=True)
+        if not config.getboolean('queue', 'worker', default=False):
+            time.sleep(0.02 * retry)
         try:
             with Transaction().start(pool.database_name, 0) as transaction:
                 task = Queue(task_id)
@@ -123,6 +135,6 @@ def run_task(pool, task_id):
                 Queue.push(task.name, task.data, scheduled_at=scheduled_at)
         except Exception:
             logger.critical(
-                'rescheduling task "%d" failed', task_id, exc_info=True)
+                'rescheduling %s failed', name, exc_info=True)
     except Exception:
-        logger.critical('task "%d" failed', task_id, exc_info=True)
+        logger.critical('%s failed', name, exc_info=True)
